# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

###

from .... import package_name
from ....utils_xslt import xslt_is_member

cibprelude2cibcompact = ('''\
    <!--
        SIMPLIFY FENCING/STONITH
     -->

    <xsl:template match="primitive[
                             @template
                             =
                             preceding-sibling::template[
                                @class = 'stonith'
                             ]/@id
                             and not(preceding-sibling::primitive[
                                @template = current()/@template
                             ])
                         ]">
        <xsl:variable name="Primitives"
                      select="(following-sibling::primitive|self::primitive)[
                                  @template = current()/@template
                              ]"/>
        <xsl:variable name="Template"
                      select="preceding-sibling::template[
                                 @id = current()/@template
                              ]"/>
        <xsl:variable name="GroupablePrimitives"
                      select="(following-sibling::primitive|self::primitive)[
                                  @template = current()/@template
                                  and not(*[name() != 'instance_attributes'])
                                  and not(instance_attributes/nvpair[
                                      @name != 'pcmk_host_list'
                                      and
                                      @name != 'pcmk_host_map'
                                  ])
                              ]"/>
        <xsl:choose>
            <!-- when we can actually do any kind of simplification/grouping,
                 i.e., when fencedev:fenceinst 1:1, or 1:N and no fenceinst
                 specify instance parameter other than port (if any)
                 or no instance attributes specified per fencedev -->
            <xsl:when test="count($Primitives) = 1
                            or
                            count($Primitives) = count($GroupablePrimitives)
                            or
                            count($Template/instance_attributes/*) = 0">
                <xsl:copy>
                    <xsl:copy-of select="$Template/@*"/>
                    <xsl:for-each select="($Primitives|$Template)/*">
                        <xsl:choose>
                            <xsl:when test="generate-id(..) = generate-id($Template)
                                            and
                                            not(
                                                $Primitives/*[
                                                    name() = name(current())
                                                ]
                                            )">
                                <xsl:copy/>
                            </xsl:when>
                            <xsl:when test="generate-id(..) = generate-id($Primitives)
                                            and name() = 'instance_attributes'">
                                <xsl:copy>
                                    <xsl:attribute name="id">
                                        <xsl:value-of select="concat($Template/@id, '-ATTRS')"/>
                                    </xsl:attribute>
                                    <xsl:copy-of select="$Template/*[
                                                             name() = name(current())
                                                         ]/*"/>
                                    <xsl:variable name="PcmkHostList"
                                                  select="$Primitives/instance_attributes/nvpair[
                                                              @name = 'pcmk_host_list'
                                                          ]"/>
                                    <xsl:variable name="PcmkHostMap"
                                                  select="$Primitives/instance_attributes/nvpair[
                                                              @name = 'pcmk_host_map'
                                                          ]"/>
                                    <xsl:if test="$PcmkHostList">
                                        <nvpair id="{concat($Template/@id, '-ATTRS-pcmk_host_list')}"
                                                name="pcmk_host_list">
                                        <xsl:attribute name="value">
                                            <xsl:for-each select="$PcmkHostList">
                                                <xsl:if test="position() != 1">
                                                    <xsl:value-of select="' '"/>
                                                </xsl:if>
                                                <xsl:value-of select="@value"/>
                                            </xsl:for-each>
                                        </xsl:attribute>
                                        </nvpair>
                                    </xsl:if>
                                    <xsl:if test="$PcmkHostMap">
                                        <nvpair id="{concat($Template/@id, '-ATTRS-pcmk_host_map')}"
                                                name="pcmk_host_map">
                                        <xsl:attribute name="value">
                                            <xsl:for-each select="$PcmkHostMap">
                                                <xsl:if test="position() != 1">
                                                    <xsl:value-of select="','"/>
                                                </xsl:if>
                                                <xsl:value-of select="@value"/>
                                            </xsl:for-each>
                                        </xsl:attribute>
                                        </nvpair>
                                    </xsl:if>
                                </xsl:copy>
                            </xsl:when>
                        </xsl:choose>
                    </xsl:for-each>
                </xsl:copy>
            </xsl:when>
            <xsl:otherwise>
                <xsl:if test="count($Primitives) != count($GroupablePrimitives)">
                    <xsl:copy-of select="preceding-sibling::template[
                                             @class = 'stonith'
                                             and
                                             @id = current()/@template
                                         ]|(following-sibling::primitive|self::primitive)[
                                             @template = current()/@template
                                         ]"/>
                </xsl:if>
            </xsl:otherwise>
        </xsl:choose>
    </xsl:template>

    <!-- remove non-first instances of fence devices, of which all are
         removed unconditionally (preserved in the above logic when needed) -->
    <xsl:template match="primitive[
                             @template
                             =
                             preceding-sibling::template[
                                @class = 'stonith'
                             ]/@id
                             and preceding-sibling::primitive[
                                @template = current()/@template
                             ]
                         ]"/>
    <xsl:template match="template[@class = 'stonith']"/>


    <!--
        trivial conversion of resource groups into groups (original groups
        are preserved as-were) if they are not exclusive
     -->

    <xsl:template match="template[
                             @provider = '%(package_name)s'
                             and
                             @type = 'temporary-service'
                             and
                             not(
                                 meta_attributes/nvpair[
                                     @name = 'exclusive'
                                     and
                                     (
                                        @value = 'yes'
                                        or
                                        @value &gt; 0
                                    )
                                 ]
                             )
                          ]">
        <xsl:variable name="ResourceGroup" select="@id"/>
        <xsl:variable name="Resources" select="../primitive[
                                                  meta_attributes/nvpair[
                                                      @name = 'rgmanager-service'
                                                      and
                                                      @value = $ResourceGroup
                                                  ]
                                               ]"/>
        <xsl:if test="$Resources">
            <group id="{$ResourceGroup}-GROUP">
                <xsl:for-each select="$Resources">
                    <xsl:copy>
                        <xsl:copy-of select="@*"/>
                        <xsl:for-each select="node()[not(
                                                name() = 'meta_attributes'
                                                and
                                                (
                                                    count(*) = 0
                                                    or
                                                    (
                                                        count(*) = 1
                                                        and
                                                        nvpair[
                                                            @name = 'rgmanager-service'
                                                        ]
                                                    )
                                                )
                                            )]">
                            <xsl:choose>
                                <xsl:when test="name() = 'meta_attributes'">
                                    <xsl:copy>
                                        <xsl:copy-of select="@*|node()[
                                                                name() != 'nvpair'
                                                                or
                                                                (
                                                                    name() = 'nvpair'
                                                                    and
                                                                    @name != 'rgmanager-service'
                                                                )]"/>
                                    </xsl:copy>
                                </xsl:when>
                                <xsl:otherwise>
                                    <xsl:copy>
                                        <xsl:copy-of select="@*|node()"/>
                                    </xsl:copy>
                                </xsl:otherwise>
                            </xsl:choose>
                        </xsl:for-each>
                    </xsl:copy>
                </xsl:for-each>
                <!--
                    stickiness=INFINITY for each N in dedicated nodes ~ @nofailback
                -->
                <xsl:variable name="FailoverDomain"
                            select="../template[
                                    @provider = '%(package_name)s'
                                    and
                                    @type = 'temporary-failoverdomain'
                                    and
                                    @id = current()/meta_attributes/nvpair[
                                        @name = 'domain'
                                    ]/@value
                                ]"/>
                <xsl:if test="$FailoverDomain/meta_attributes/nvpair[
                                @name = 'nofailback'
                            ]/@value ='1'
                            and
                            count(
                                $FailoverDomain/meta_attributes/nvpair[
                                    starts-with(@name, 'failoverdomainnode-')
                                ]
                            ) != 0">
                    <xsl:comment
                    ><xsl:value-of select="concat(' mimic NOFAILBACK failoverdomain (',
                                                $FailoverDomain/@id, ')')"
                    /></xsl:comment>
                    <meta_attributes id="{$ResourceGroup}-META-ATTRS-nofailback">
                        <rule id="{$ResourceGroup}-META-RULE-stickiness"
                            score="0"
                            boolean-op="or">
                            <xsl:for-each select="$FailoverDomain/meta_attributes/nvpair[
                                                    starts-with(@name, 'failoverdomainnode-')
                                                ]">
                                <expression id="STICKINESS-{$ResourceGroup}-{@value}"
                                            attribute="#uname"
                                            operation="eq"
                                            value="{@value}">
                                </expression>
                            </xsl:for-each>
                        </rule>
                        <nvpair id="{$ResourceGroup}-META-ATTRS-nofailback-pair"
                                name="resource-stickiness"
                                value="INFINITY"/>
                    </meta_attributes>
                </xsl:if>
                <!--
                    is-managed=false ~ @autostart in (no, 0)
                -->
                <xsl:variable name="Autostart"
                            select="meta_attributes/nvpair[
                                        @name = 'autostart'
                                    ]/@value"/>
                <xsl:if test="$Autostart = 'no'
                              or
                              $Autostart = 0">
                    <xsl:comment
                    ><xsl:value-of select="concat(' mimic no-autostart resource group (',
                                                  $ResourceGroup, ')')"
                    /></xsl:comment>
                    <meta_attributes id="{$ResourceGroup}-META-ATTRS-autostart">
                        <nvpair id="{$ResourceGroup}-META-is-managed"
                                name="is-managed"
                                value="false"/>
                    </meta_attributes>
                </xsl:if>
                <!--
                    migration-treshold + failure-timeout
                    ~ @max_restarts + restart_expire_time
                      iff recovery in (restart, relocate->migration-threshold=1)
                -->
                <xsl:variable name="MaxRestarts"
                            select="meta_attributes/nvpair[
                                        @name = 'max_restarts'
                                    ]/@value"/>
                <xsl:variable name="RestartExpireTime"
                            select="meta_attributes/nvpair[
                                        @name = 'restart_expire_time'
                                    ]/@value"/>
                <xsl:variable name="Recovery"
                            select="translate(
                                        meta_attributes/nvpair[
                                            @name = 'recovery'
                                        ]/@value,
                                        'ABCDEFGHIJKLMNOPQRSTUVWXYZ',
                                        'abcdefghijklmnopqrstuvwxyz'
                                    )"/>
                <xsl:if test="
''' + (
                    xslt_is_member('$Recovery',
                                   ('', 'restart', 'relocate'))
) + '''
                ">
                    <xsl:choose>
                        <xsl:when test="$Recovery = 'relocate'">
                            <xsl:comment
                            ><xsl:value-of select="concat(
                                ' mimic relocate recover policy for',
                                ' resource group (', $ResourceGroup, ')')"
                            /></xsl:comment>
                            <meta_attributes id="{$ResourceGroup}-META-ATTRS-recover-relocate">
                                <nvpair id="{$ResourceGroup}-META-migration-threshold"
                                        name="migration-threshold"
                                        value="1"/>
                                <!-- 30s timeout is non-educated guestimation -->
                                <nvpair id="{$ResourceGroup}-META-failure-timeout"
                                        name="failure-timeout"
                                        value="30s"/>
                            </meta_attributes>
                        </xsl:when>
                        <xsl:when test="number($MaxRestarts) &gt; 0
                                        and
                                        number($RestartExpireTime) &gt;= 0">
                            <xsl:comment
                            ><xsl:value-of select="concat(
                                ' mimic finely specified restart recover',
                                ' policy for resource group (',
                                $ResourceGroup, ')')"
                            /></xsl:comment>
                            <meta_attributes id="{$ResourceGroup}-META-ATTRS-recover-restart">
                                <nvpair id="{$ResourceGroup}-META-migration-threshold"
                                        name="migration-threshold"
                                        value="{$MaxRestarts}"/>
                                <!-- 30s timeout is non-educated guestimation -->
                                <nvpair id="{$ResourceGroup}-META-failure-timeout"
                                        name="failure-timeout"
                                        value="{$RestartExpireTime}"/>
                            </meta_attributes>
                        </xsl:when>
                        <xsl:otherwise>
                        </xsl:otherwise>
                    </xsl:choose>
                </xsl:if>
            </group>
        </xsl:if>

        <xsl:if test="meta_attributes/nvpair[
                          @name = 'domain'
                      ]">
            <xsl:copy>
                <xsl:copy-of select="@*|node()"/>
            </xsl:copy>
        </xsl:if>
    </xsl:template>

    <!-- also remove the primitive(s) now moved to the group(s) -->
    <xsl:template match="primitive[
                            meta_attributes/nvpair[
                                @name = 'rgmanager-service'
                                and
                                @value = current()/../template[
                                    @provider = '%(package_name)s'
                                    and
                                    @type = 'temporary-service'
                                    and
                                    not(
                                        meta_attributes/nvpair[
                                            @name = 'exclusive'
                                            and
                                            (
                                               @value = 'yes'
                                               or
                                               @value &gt; 0
                                           )
                                        ]
                                    )
                                ]/@id
                            ]
                         ]"/>

    <!-- exclusive resource groups (pre-conversion state) not supported yet -->
    <xsl:template match="template[
                             @provider = '%(package_name)s'
                             and
                             @type = 'temporary-service'
                             and
                             meta_attributes/nvpair[
                                 @name = 'exclusive'
                                 and
                                 (
                                    @value = 'yes'
                                    or
                                    @value &gt; 0
                                )
                             ]
                          ]">
        <xsl:message terminate="yes">
            <xsl:value-of select="concat('Cannot convert resource group',
                                         ' when exclusive (could be mimicked',
                                         ' using utilization properties or',
                                         ' highly unscalable set of',
                                         ' anti-colocations): ',
                                         substring-after(@id, 'SERVICE-'),
                                         ' [https://bugzilla.redhat.com/1206640#c2]')"/>
        </xsl:message>
    </xsl:template>
''') % dict(package_name=package_name())

###

from ....filters._2pcscmd import verbose_ec_test, verbose_inform
from ....filters.cib2pcscmd import attrset_xsl
from ....utils_xslt import NL

cib2pcscmd = ('''\
    <!-- STONITH -->
    <xsl:for-each select=".//primitive[@class = 'stonith']">
''' + (
        verbose_inform('"new stonith: ", @id')
) + '''
        <xsl:variable name="PcsStonithCmd">
            <xsl:choose>
            <!-- "pcs stonith create ID external/EXT" only supported
                 with certain newer versions of pcs
                 (https://github.com/ClusterLabs/pcs/issues/81) -->
                <xsl:when test="$pcscmd_extra_agents_via_pacemaker
                                or
                                not(starts-with(@type, 'external/'))">
                    <xsl:value-of select="$pcscmd_pcs"/>
                </xsl:when>
                <xsl:otherwise>
                    <xsl:message>%(stonith_external_msg)s</xsl:message>
                    <xsl:value-of select="concat($pcscmd_pcs, '--force ')"/>
                </xsl:otherwise>
            </xsl:choose>
        </xsl:variable>
        <xsl:value-of select="concat($PcsStonithCmd, 'stonith create',
                                     ' ', @id,
                                     ' ', @type)"/>
''' + (
        attrset_xsl("instance_attributes")
) + '''
        <!-- operations -->
        <xsl:if test="operations/op">
            <xsl:value-of select="' op'"/>
            <xsl:for-each select="operations/op">
                <xsl:value-of select="concat(' ', @name)"/>
                <xsl:for-each select="@*">
                    <xsl:value-of select='concat(" &apos;",
                                                 name(), "=", .,
                                                 "&apos;")'/>
                </xsl:for-each>
            </xsl:for-each>
        </xsl:if>
        <!-- meta attrs -->
''' + (
        attrset_xsl("meta_attributes", cmd='" meta"')
) + '''
        <xsl:value-of select="'%(NL)s'"/>
''' + (
        verbose_ec_test
) + '''
    </xsl:for-each>

    <!--
        ORDINARY/CLONE/MASTER CLUSTER RESOURCES
     -->

    <!-- bundles -->
    <clufter:descent-mix at="bundle"/>

    <!-- primitives (can depend on bundle, but only if supported) -->
    <xsl:for-each select=".//primitive[@class != 'stonith'
                                       and
                                       (
                                           name(..) != 'bundle'
                                           or
                                           $pcscmd_extra_bundle
                                       )]">
        <xsl:variable name="ResourceSpec">
            <xsl:choose>
                <xsl:when test="@class = 'ocf'">
                    <xsl:value-of select="concat(@class, ':', @provider, ':', @type)"/>
                </xsl:when>
                <xsl:when test="@class">
                    <xsl:value-of select="concat(@class, ':', @type)"/>
                </xsl:when>
                <xsl:otherwise>
                    <xsl:value-of select="@type"/>
                </xsl:otherwise>
            </xsl:choose>
        </xsl:variable>
        <xsl:choose>
            <xsl:when test="name(..) = 'clone'">
''' + (
                verbose_inform('"new resource: ", @id, " (to be set as clone)"')
) + '''
            </xsl:when>
            <xsl:when test="name(..) = 'master'">
''' + (
                verbose_inform('"new resource: ", @id, " (to be set as master)"')
) + '''
            </xsl:when>
            <xsl:when test="name(..) = 'bundle'">
''' + (
                verbose_inform('"new resource: ", @id, " (being added to bundle)"')
) + '''
            </xsl:when>
            <xsl:otherwise>
''' + (
                verbose_inform('"new resource: ", @id')
) + '''
            </xsl:otherwise>
        </xsl:choose>
        <xsl:value-of select="concat($pcscmd_pcs, 'resource create',
                                     ' ', @id,
                                     ' ', $ResourceSpec)"/>
''' + (
        attrset_xsl("instance_attributes")
) + '''
        <!-- operations -->
        <xsl:if test="operations/op">
            <xsl:value-of select="' op'"/>
            <xsl:for-each select="operations/op">
                <xsl:value-of select="concat(' ', @name)"/>
                <xsl:for-each select="@*[
                                          name() != 'id'
                                          and
                                          name() != 'name'
                                      ]">
                    <xsl:value-of select='concat(" &apos;",
                                                 name(), "=", .,
                                                 "&apos;")'/>
                </xsl:for-each>
            </xsl:for-each>
        </xsl:if>
        <!-- meta attrs -->
''' + (
        attrset_xsl("meta_attributes", cmd='" meta"')
) + '''
        <!-- possible containment within (existing) bundle -->
        <xsl:if test="name(..) = 'bundle'">
            <xsl:value-of select="concat(' bundle', ' ', ../@id)"/>
        </xsl:if>

        <!-- NOTE clone/master resource specifics handled separately later -->
        <xsl:value-of select="'%(NL)s'"/>
''' + (
        verbose_ec_test
) + '''

        <!-- "pcs resource utilization" only supported with certain newer
             versions of pcs (https://bugzilla.redhat.com/1158500) -->
        <xsl:choose>
            <xsl:when test="$pcscmd_extra_utilization">
''' + (
            attrset_xsl("utilization",
                        cmd='$pcscmd_pcs, "resource utilization ", @id',
                        inform='"set utilization for resource: ", @id')
) + '''
            </xsl:when>
            <xsl:when test="utilization/nvpair">
                <xsl:message>%(utilization_msg)s</xsl:message>
            </xsl:when>
        </xsl:choose>

    </xsl:for-each>

    <!-- group -->
    <clufter:descent-mix at="group"/>

    <!-- clone (depends on primitive|group) -->
    <clufter:descent-mix at="clone"/>

    <!-- master (depends on primitive|group) -->
    <clufter:descent-mix at="master"/>

    <!-- templates -->
    <xsl:if test="template">
        <xsl:message terminate="yes"
        >Cannot convert templates to pcs commands yet [https://bugzilla.redhat.com/1281359]</xsl:message>
    </xsl:if>

''') % dict(
    NL=NL,
    utilization_msg="WARNING: target pcs version does not support utilization"
                    " attributes, hence omitted",
    stonith_external_msg="WARNING: target pcs version does not support"
                         " getting agents metadata through pacemaker, hence"
                         " special stonith case of external/* needs to be"
                         " passed by force",
)

###

from ....utils_xslt import xslt_is_member

cib_meld_templates_op_roles = ('Started', 'Slave')

# see lib/pengine/complex.c: unpack_template
cib_meld_templates = ('''\
    <!-- drop any occurrence as we meld them into proper primitives -->
    <xsl:template match="template[1]">
        <xsl:message
        >NOTE: this step may not be necessary once pcs handles templates [https://bugzilla.redhat.com/1281359]</xsl:message>
    </xsl:template>
    <xsl:template match="template"/>

    <xsl:template name="rewrite-id">
        <xsl:param name="Elem"/>
        <xsl:param name="InstanceId"/>
        <xsl:copy>
            <xsl:for-each select="@*">
                <xsl:choose>
                    <xsl:when test="name() = 'id'">
                        <xsl:attribute name="{name()}">
                            <xsl:value-of select="concat(., '-', $InstanceId)"/>
                        </xsl:attribute>
                    </xsl:when>
                    <xsl:otherwise>
                        <xsl:copy/>
                    </xsl:otherwise>
                </xsl:choose>
            </xsl:for-each>
            <xsl:for-each select="*">
                <xsl:call-template name="rewrite-id">
                    <xsl:with-param name="Elem" select="."/>
                    <xsl:with-param name="InstanceId" select="$InstanceId"/>
                </xsl:call-template>
            </xsl:for-each>
        </xsl:copy>
    </xsl:template>

    <xsl:template match="primitive[@template]">
        <xsl:variable name="Template"
                      select="//template[@id = current()/@template]"/>
        <xsl:variable name="InstanceId" select="generate-id()"/>
        <xsl:copy>
            <xsl:copy-of select="@id|$Template/@*[name() != 'id']"/>
            <xsl:for-each select="$Template/*[name() != 'operations']">
                <xsl:call-template name="rewrite-id">
                    <xsl:with-param name="Elem" select="."/>
                    <xsl:with-param name="InstanceId" select="$InstanceId"/>
                </xsl:call-template>
            </xsl:for-each>
            <xsl:for-each select="*">
                <xsl:copy>
                    <xsl:copy-of select="@*"/>
                    <xsl:if test="name() = 'operations'">
                        <xsl:variable name="Operations" select="."/>
                        <xsl:for-each select="$Template/op[
                                                  not(
                                                      $Operations/op[
                                                          @name = current()/@name
                                                          and
                                                          (
                                                              @role = current()/@role
                                                              or
                                                              ((
                                                                  not(@role)
                                                                  or
''' + (
                                                                  xslt_is_member('@role',
                                                                                 cib_meld_templates_op_roles)
) + '''
                                                              ) and (
                                                                  not(current()/@role)
                                                                  or
''' + (
                                                                  xslt_is_member('current()/@role',
                                                                                 cib_meld_templates_op_roles)
) + '''

                                                              ))
                                                          )
                                                      ]
                                                  )
                                              ]">
                            <xsl:call-template name="rewrite-id">
                                <xsl:with-param name="Elem" select="."/>
                                <xsl:with-param name="InstanceId" select="$InstanceId"/>
                            </xsl:call-template>
                        </xsl:for-each>
                    </xsl:if>
                    <xsl:if test="name() != 'operations'">
                        <xsl:apply-templates/>
                    </xsl:if>
                </xsl:copy>
            </xsl:for-each>
        </xsl:copy>
    </xsl:template>
''')

###

cibcompact2cib = ('''\
    <!-- propagate monitor/independent tree info to proper operations stanza -->
    <xsl:template match="primitive[
                             meta_attributes/nvpair[
                                 @name = 'rgmanager-monitor'
                                 or
                                 @name = 'rgmanager-independent'
                             ]
                          ]">
        <xsl:variable name="ResPrefix"
                      select="@id"/>
        <xsl:variable name="Monitor"
                      select="meta_attributes/nvpair[
                                  @name = 'rgmanager-monitor'
                              ]"/>
        <xsl:variable name="IsIndependent"
                      select="meta_attributes/nvpair[
                                  @name = 'rgmanager-independent'
                                  and
                                  @value = '2'
                              ]"/>
        <xsl:copy>
            <xsl:copy-of select="@*"/>
            <xsl:for-each select="*[name() != 'utilization']">
                <xsl:choose>
                    <xsl:when test="name() = 'meta_attributes'
                                    and
                                    count(*) = count(
                                        nvpair[
                                            @name = 'rgmanager-monitor'
                                            or
                                            @name = 'rgmanager-independent'
                                        ]
                                    )"/>
                    <xsl:when test="name() = 'meta_attributes'">
                        <xsl:copy>
                            <xsl:apply-templates select="@*|*[
                                                             name() != 'nvpair'
                                                             or
                                                             @name != 'rgmanager-monitor'
                                                             or
                                                             @name != 'rgmanager-independent'
                                                         ]"/>
                        </xsl:copy>
                    </xsl:when>
                    <xsl:when test="name() = 'operations'">
                        <xsl:copy>
                            <xsl:copy-of select="@*"/>
                            <xsl:for-each select="op">
                                <xsl:copy>
                                    <xsl:copy-of select="@*"/>
                                    <xsl:choose>
                                        <xsl:when test="(
                                                            @name = 'monitor'
                                                            or
                                                            @name = 'status'
                                                        )
                                                        and
                                                        $IsIndependent">
                                            <xsl:attribute name="on-fail">
                                                <xsl:value-of select="'ignore'"/>
                                            </xsl:attribute>
                                        </xsl:when>
                                        <xsl:when test="@name = 'stop'
                                                        and
                                                        $IsIndependent">
                                            <xsl:attribute name="on-fail">
                                                <xsl:value-of select="'stop'"/>
                                            </xsl:attribute>
                                        </xsl:when>
                                    </xsl:choose>
                                </xsl:copy>
                            </xsl:for-each>
                            <xsl:if test="not(
                                              op[
                                                  @name = 'monitor'
                                                  or
                                                  @name = 'status'
                                              ]
                                          )
                                          and
                                          (
                                              $Monitor
                                              or
                                              $IsIndependent
                                          )">
                                <!-- interval has to be defined -->
                                <op id="{concat($ResPrefix, '-OP-monitor')}"
                                    name="monitor"
                                    interval="60s">
                                    <xsl:if test="$Monitor">
                                        <xsl:attribute name="interval">
                                            <xsl:value-of select="$Monitor/@value"/>
                                        </xsl:attribute>
                                    </xsl:if>
                                    <xsl:if test="$IsIndependent">
                                        <xsl:attribute name="on-fail">
                                            <xsl:value-of select="'ignore'"/>
                                        </xsl:attribute>
                                    </xsl:if>
                                </op>
                            </xsl:if>
                            <xsl:if test="not(
                                              op[@name = 'stop']
                                          )
                                          and
                                          $IsIndependent">
                                <op id="{concat($ResPrefix, '-OP-stop')}"
                                    name="stop"
                                    interval="0"
                                    on-fail="stop"/>
                            </xsl:if>
                        </xsl:copy>
                    </xsl:when>
                    <xsl:otherwise>
                        <xsl:copy>
                            <xsl:apply-templates select="@*|*"/>
                        </xsl:copy>
                    </xsl:otherwise>
                </xsl:choose>
            </xsl:for-each>
            <xsl:if test="not(operations)
                          and
                          (
                              $Monitor
                              or
                              $IsIndependent
                          )">
                <operations>
                    <!-- interval has to be defined -->
                    <op id="{concat($ResPrefix, '-OP-monitor')}"
                        name="monitor"
                        interval="60s">
                        <xsl:if test="$Monitor">
                            <xsl:attribute name="interval">
                                <xsl:value-of select="$Monitor/@value"/>
                            </xsl:attribute>
                        </xsl:if>
                        <xsl:if test="$IsIndependent">
                            <xsl:attribute name="on-fail">
                                <xsl:value-of select="'ignore'"/>
                            </xsl:attribute>
                        </xsl:if>
                    </op>
                    <xsl:if test="$IsIndependent">
                        <op id="{concat($ResPrefix, '-OP-stop')}"
                            name="stop"
                            interval="0"
                            on-fail="stop"/>
                    </xsl:if>
                </operations>
            </xsl:if>
            <xsl:apply-templates select="utilization"/>
        </xsl:copy>
    </xsl:template>
''')



# -*- coding: UTF-8 -*-
# Copyright 2014 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""Cluster configuration system (ccs) format"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ..format import XML


class ccs(XML):
    """CMAN-based cluster stack configuration (cluster.conf)

    Sometimes called Cluster Configuration System (ccs).
    """
    # XML
    root = 'cluster'
    validator_specs = {
        XML.ETREE: ''  # XXX no RNG schema handy yet
    }


class ccs_flat(ccs):
    """Private, artificially flattened CMAN-based cluster stack configuration

    This is a result of (one-off!) linearization of nested explicit ordering
    of resources as performed internally by RGManager, and in turn, also
    by the derived helper ccs_flatten (bundled) ... wrapped in ccs2ccsflat
    filter.
    """



# -*- coding: UTF-8 -*-
# Copyright 2015 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)

from ....utils_xslt import xslt_is_member

###

ccsflat2cibprelude = '''\
    <template id="{concat('FENCEDEV-', @name)}"
              class="stonith"
              type="{@agent}">
        <xsl:variable name='FenceDevName' select="@name"/>
        <instance_attributes id="{concat('FENCEDEV-', @name, '-ATTRS')}">
        <xsl:for-each select="@*[name() != 'agent' and name() != 'name']">
            <nvpair id="{concat('FENCEDEV-', $FenceDevName, '-ATTRS-', name())}"
                    name="{name()}"
                    value="{.}"/>
        </xsl:for-each>
        </instance_attributes>
    </template>
'''

###

ccs_obfuscate_credentials_password = (
    'passwd',
    'snmp_priv_passwd',
    # legacy fence_vmware https://fedorahosted.org/cluster/wiki/VMware_FencingConfig#OldFence_vmware
    'vmpasswd',
)

ccs_obfuscate_credentials_login = (
    'login',
)

ccs_obfuscate_credentials = '''\
    <xsl:copy>
        <xsl:apply-templates select="@*"/>
        <xsl:for-each select="@*[
''' + (
        xslt_is_member('name()', ccs_obfuscate_credentials_password)
) + ''']">
            <xsl:attribute name="{name()}">SECRET-PASSWORD</xsl:attribute>
        </xsl:for-each>
        <xsl:for-each select="@*[
''' + (
        xslt_is_member('name()', ccs_obfuscate_credentials_login)
) + ''']">
            <xsl:attribute name="{name()}">SECRET-LOGIN</xsl:attribute>
        </xsl:for-each>
        <xsl:apply-templates/>
    </xsl:copy>
'''

###

ccs_revitalize = '''\
    <!-- xvm: domain -> port -->
    <xsl:template match="fencedevice[@agent = 'fence_xvm']">
        <xsl:copy>
            <xsl:for-each select="@*">
                <xsl:choose>
                    <xsl:when test="name() = 'domain'">
                        <xsl:attribute name='port'>
                            <xsl:value-of select="."/>
                        </xsl:attribute>
                    </xsl:when>
                    <xsl:otherwise>
                        <xsl:copy/>
                    </xsl:otherwise>
                </xsl:choose>
            </xsl:for-each>
            <xsl:apply-templates select="*"/>
        </xsl:copy>
    </xsl:template>
'''

###

from ....filters.ccs_artefacts import artefact, artefact_cond

ccs_artefacts_common_params = (
    # #el6        # el7
    # alom        alom
    # apc         apc
    # bladecenter bladecenter
    #             brocade
    # drac5       drac5
    #             hds_cb
    # hpblade     hpblade
    # ilo_mp      ilo_mp
    # ldom        ldom
    # lpar        lpar
    # rsa         rsa
    # rsb         rsb
    # virsh       virsh
    # vmware      vmware
    # wti         wti
    artefact_cond('@identity_file', kind='S',
                  desc='FA: SSH identity file (private key)')
    ) + (
    # #el6         #el7
    # alom         alom
    # apc          apc
    # apc_snmp     apc_snmp
    # baytech      baytech
    # bladecenter  bladecenter
    # brocade      brocade
    # bullpap      bullpap
    # cisco_mds    cisco_mds
    # cisco_ucs    cisco_ucs
    # drac         drac
    # drac5        drac5
    # eaton_snmp   eaton_snmp
    # eps          eps
    #              hds_cb
    # hpblade      hpblade
    # ibmblade     ibmblade
    # ifmib        ifmib
    # ilo          ilo
    # ilo_mp       ilo_mp
    # intelmodular intelmodular
    # ipdu         ipdu
    # ipmilan      ipmilan
    # ldom         ldom
    # lpar         lpar
    # mcdata       mcdata
    #              netio
    #              ovh
    # rackswitch   rackswitch
    # rhevm        rhevm
    # rsa          rsa
    # rsb          rsb
    # sanbox2      sanbox2
    # virsh        virsh
    # vixel        vixel
    # vmware       vmware
    # vmware_soap  vmware_soap
    # wti          wti
    # xenapi       xenapi
    # zvm          zvm
    artefact_cond('@passwd_script', kind='S',
                  desc='FA: script to retrieve password')
    ) + (
    #  #el6          #el7
    #  alom          alom
    #  apc           apc
    #  apc_snmp      apc_snmp
    #  bladecenter   bladecenter
    #                brocade
    #   cisco_mds    cisco_mds
    #   cisco_ucs    cisco_ucs
    #   drac         drac
    #   drac5        drac5
    #                dummy
    #   eaton_snmp   eaton_snmp
    #   eps          eps
    #                hds_cb
    #   hpblade      hpblade
    #   ibmblade     ibmblade
    #   ifmib        ifmib
    #   ilo          ilo
    #   ilo_mp       ilo_mp
    #   intelmodular intelmodular
    #   ipdu         ipdu
    #   ldom         ldom
    #   lpar         lpar
    #                netio
    #                ovh
    #   rhevm        rhevm
    #   rsa          rsa
    #   rsb          rsb
    #   sanbox2      sanbox2
    #   virsh        virsh
    #   virt         virt
    #   vmware       vmware
    #   vmware_soap  vmware_soap
    #   wti          wti
    #   xenapi       xenapi
    #   xvm          xvm
    artefact_cond('@debug', kind='L',
                  desc='FA: log file with debug information')
    ) + (
    # #el6      #el7
    # virt      virt
    # xvm       xvm
    artefact_cond('@key_file', kind='S',
                  desc='FA: shared key file')
    )

ccs_artefacts = artefact("concat('/usr/sbin/', @agent)", kind='X',
                         desc='fence agent') + ccs_artefacts_common_params



# -*- coding: UTF-8 -*-
# Copyright 2015 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

###

from .... import package_name

cib2cibfinal = ('''\
    <xsl:choose>
        <xsl:when test="not(@provider = '%(package_name)s'
                            and
                            starts-with(@type, 'temporary-'))">
            <xsl:copy>
                <xsl:copy-of select="@*|node()"/>
            </xsl:copy>
        </xsl:when>
    </xsl:choose>
''') % dict(package_name=package_name())



# -*- coding: UTF-8 -*-
# Copyright 2016 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ....filters._2pcscmd import verbose_ec_test, verbose_inform
from ....utils_xslt import NL, translate_lower, xslt_is_member


cib2pcscmd_options = (
    'score',
    # following are highly questionable:
    # https://github.com/ClusterLabs/pacemaker/pull/1091
    #'score-attribute',
    #'score-attribute-mangle',
)

cib2pcscmd_set_options = (
    'role',
    'sequential',
    'score',
    'ordering',  # since constraints-2.1.rng
)

cib2pcscmd = ('''\
    <xsl:choose>
        <!-- plain "A with B" -->
        <xsl:when test="@rsc and @with-rsc">
''' + (
            verbose_inform('"new colocation constraint (resource pair): ", @id')
) + '''
            <xsl:value-of select="concat($pcscmd_pcs, 'constraint colocation add')"/>
            <xsl:if test="
''' + (
            xslt_is_member(translate_lower('@rsc-role'),
                           ('master', 'slave'))
) + '''">
                <xsl:value-of select="concat(' ',
''' + (
                    translate_lower('@rsc-role')
) + ''')"/>
            </xsl:if>
            <xsl:value-of select="concat(' ', @rsc, ' with')"/>
            <xsl:if test="
''' + (
            xslt_is_member(translate_lower('@with-rsc-role'),
                           ('master', 'slave'))
) + '''">
                <xsl:value-of select="concat(' ',
''' + (
                    translate_lower('@with-rsc-role')
) + ''')"/>
            </xsl:if>
            <xsl:value-of select="concat(' ', @with-rsc)"/>
            <xsl:choose>
                <!-- missing score is considered INFINITY by pcs anyway -->
                <xsl:when test="@score and not(
''' + (
                    xslt_is_member('@score', ('INFINITY', '+INFINITY'))
) + ''')">
                    <xsl:value-of select="concat(' ', @score)"/>
                </xsl:when>
                <xsl:when test="@score"/>
                <xsl:otherwise>
                    <!-- missing score would be considered INFINITY by pcs -->
                    <xsl:value-of select="' 0'"/>
                </xsl:otherwise>
            </xsl:choose>
            <xsl:value-of select="concat(' ', 'id=', @id)"/>
            <xsl:value-of select="'%(NL)s'"/>
''' + (
            verbose_ec_test
) + '''
        </xsl:when>

        <!-- resource sets -->
        <xsl:when test="resource_set">
''' + (
            verbose_inform('"new colocation constraint (resource set): ", @id')
) + '''
            <xsl:value-of select="concat($pcscmd_pcs, 'constraint colocation')"/>

            <xsl:for-each select="resource_set">
                <xsl:value-of select="' set'"/>
                <xsl:for-each select="resource_ref">
                    <xsl:value-of select="concat(' ', @id)"/>
                </xsl:for-each>

                <xsl:for-each select="@*[
''' + (
                    xslt_is_member('name()', cib2pcscmd_set_options)
) + ''']">
                    <xsl:value-of select="concat(' ', name(), '=', .)"/>
                </xsl:for-each>

            </xsl:for-each>

            <xsl:value-of select="' setoptions'"/>
            <xsl:for-each select="@*[
''' + (
                xslt_is_member('name()', cib2pcscmd_options)
) + ''']">
                <xsl:value-of select="concat(' ', name(), '=', .)"/>
            </xsl:for-each>
            <xsl:value-of select="concat(' ', 'id=', @id)"/>
            <xsl:value-of select="'%(NL)s'"/>
''' + (
            verbose_ec_test
) + '''
        </xsl:when>
    </xsl:choose>

''') % dict(
    NL=NL,
)



# -*- coding: UTF-8 -*-
# Copyright 2014 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)

from ....utils_xslt import xslt_is_member

###

ccs2needlexml_attrs = (
    'debug',
    'logfile',
    'logfile_priority',
    'syslog_facility',
    'syslog_priority',
    'to_logfile',
    'to_syslog',
)

ccs2needlexml = '''\
    <logging>
        <xsl:copy-of select="@*[
''' + (
            xslt_is_member('name()', ccs2needlexml_attrs)
) + ''']"/>
        <!-- XXX: the latter match (if any) should overwrite the former -->
        <clufter:descent at="logging_daemon"/>
    </logging>
'''

###

from ....filters.ccs_artefacts import artefact_cond

ccs_artefacts = artefact_cond('@logfile', kind='F', desc='base logfile')



# -*- coding: UTF-8 -*-
# Copyright 2014 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)

from ....utils_xslt import xslt_is_member

###

ccs2needlexml_attrs = (
    'gid',
    'uid',
)

ccs2needlexml = '''\
    <xsl:copy>
        <xsl:copy-of select="@*[
''' + (
            xslt_is_member('name()', ccs2needlexml_attrs)
) + ''']"/>
    </xsl:copy>
'''



# -*- coding: UTF-8 -*-
# Copyright 2015 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ....utils_cib import ResourceSpec, rg2hb_xsl


ccsflat2cibprelude = '''\
    <!--
        nfsserver ~ nfsserver
     -->
    <xsl:when test="name() = 'nfsserver'">
''' + (
        ResourceSpec('ocf:heartbeat:nfsserver').xsl_attrs
) + '''
        <!-- INSTANCE_ATTRIBUTES -->
        <instance_attributes id="{concat($Prefix, '-ATTRS')}">
''' + (
            rg2hb_xsl('nfs_shared_infodir', 'nfspath')
            +
            # see rhbz#918315 + rhbz#1096376/7
            rg2hb_xsl('statd_port', 'statdport')
) + '''\
        </instance_attributes>
    </xsl:when>
'''

###

from ....filters.ccs_artefacts import artefact_cond_ra

ccs_artefacts = ''.join((
    artefact_cond_ra('@path',
                     kind='I', own=True, desc='exported path'),
))



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ....filters._2pcscmd import verbose_ec_test, verbose_inform
from ....utils_xslt import NL

cib2pcscmd = ('''\
    <!-- "pcs acl" only supported with certain newer versions of
         pcs/pacemaker (https://bugzilla.redhat.com/1111369) -->
    <xsl:choose>
        <xsl:when test="$pcscmd_extra_acls">
''' + (
            verbose_inform('"new ACL user: ", @id')
) + '''
            <xsl:value-of select="concat($pcscmd_pcs, 'acl user create', ' ', @id)"/>
            <xsl:for-each select="role">
                <xsl:value-of select="concat(' ', @id)"/>
            </xsl:for-each>
            <xsl:value-of select="'%(NL)s'"/>
''' + (
            verbose_ec_test
) + '''
        </xsl:when>
        <xsl:when test="not(preceding-sibling::acl_group)">
            <xsl:message>%(acls_msg)s</xsl:message>
        </xsl:when>
    </xsl:choose>
''') % dict(
    NL=NL,
    acls_msg="WARNING: target pacemaker/pcs versions do not support"
             " (new) ACLs, hence &apos;acl_group&apos; element(s) omitted",
)



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""Base command stuff (TBD)"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from collections import MutableMapping
try:
    from itertools import zip_longest
except ImportError:  # PY2 backward compatibility
    from itertools import izip_longest as zip_longest
from logging import getLogger
from optparse import OptionParser
from sys import stderr, stdin, stdout
from time import time

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

from .command_context import CommandContext
from .error import ClufterError, \
                   EC
from .filter import Filter, CMD_HELP_OPTSEP_COMMON
from .format import FormatError, SimpleFormat
from .plugin_registry import PluginRegistry
from .protocol import protodictval
from .utils import any2iter, \
                   areinstancesupto, \
                   args2tuple, \
                   filterdict_keep, \
                   head_tail, \
                   hybridproperty, \
                   nonetype, \
                   selfaware, \
                   tuplist
from .utils_2to3 import MimicMeta, basestring, \
                        foreach_u, filter_u, \
                        iter_items, iter_values, \
                        xrange
from .utils_func import apply_aggregation_preserving_depth, \
                        apply_intercalate, \
                        apply_loose_zip_preserving_depth, \
                        apply_preserving_depth, \
                        bifilter, \
                        foreach, \
                        tailshake, \
                        zip_empty
from .utils_prog import FancyOutput, \
                        cli_decor, \
                        longopt_letters_reprio, \
                        defer_common

log = getLogger(__name__)

fltiodecl = lambda x: len(x) == 2 and isinstance(x[0], Filter)

# expected to be lowercase for more straightforward case-insensitive comparison
CMD_HELP_OPTSEP_PRIMARY = 'options:'
CMD_HELP_OPTSEP_COMMON =  CMD_HELP_OPTSEP_COMMON.lower()


class CommandError(ClufterError):
    pass


class commands(PluginRegistry):
    """Command registry (to be used as a metaclass for commands)"""
    pass


class _Command(object):
    """Base for commands, i.e., encapsulations of filter chains

    Also see the docstring for `deco`.
    """

    @MimicMeta.classmethod
    def _resolve_filter_chain(cls, filters):
        res_input = cls._filter_chain
        res_output = apply_preserving_depth(filters.get)(res_input)
        if apply_aggregation_preserving_depth(all)(res_output):
            log.debug("resolve at `{0}' command: `{1}' -> {2}"
                      .format(cls.name, repr(res_input), repr(res_output)))
            return res_output
        # drop the command if cannot resolve any of the filters
        res_input = apply_intercalate(res_input)
        foreach_u(lambda i, x: log.warning("Resolve at `{0}' command:"
                                           " `{1}' (#{2}) filter fail"
                                           .format(cls.name, res_input[i], i)),
            filter_u(lambda i, x: not(x),
                     enumerate(apply_intercalate(res_output))))
        return None

    @MimicMeta.passdeco(hybridproperty)
    def filter_chain(this):
        """Chain of filter identifiers/classes for the command"""
        return this._filter_chain

    @MimicMeta.method
    def __new__(cls, filters, *args):
        filter_chain = cls._resolve_filter_chain(filters)
        if filter_chain is None:
            return None
        self = super(Command, cls).__new__(cls)
        self._filter_chain = filter_chain
        self._filters = OrderedDict((f.__class__.name, f) for f in
                                    apply_intercalate(filter_chain))
        fnc_defaults, fnc_varnames = self._fnc_defaults_varnames
        for varname, default in iter_items(fnc_defaults):
            if not isinstance(default, basestring):
                continue
            try:
                # early/static interpolation of defaults ~ filters' constants
                fnc_defaults[varname] = default.format(**self._filters)
            except AttributeError:
                pass
        self._fnc_defaults_varnames = fnc_defaults, fnc_varnames
        # following will all be resolved lazily, on-demand;
        # all of these could be evaluated upon instantiation immediately,
        # but this is not the right thing to do due to potentially many
        # commands being instantiated initially, while presumably only one
        # of them will be run later on
        self._desc_opts = None
        self._filter_chain_analysis = None  # will be dict
        return self

    #
    # filter chain related
    #

    @MimicMeta.passdeco(property)
    def filter_chain_analysis(self):
        if self._filter_chain_analysis is None:
            filter_chain = self._filter_chain
            self._filter_chain_analysis = self.analyse_chain(filter_chain)
        return self._filter_chain_analysis

    @MimicMeta.staticmethod
    @selfaware
    def analyse_chain(me, filter_chain, analysis_acc=None):
        """Given the filter chain, return filter backtrack and terminal chain

        This is done by recursive traversal.  Also check that the graph is
        actually connected wrt. protocols compatibility between each of
        adjacent filters is performed.

        XXX: mentioned check doesn't know about CompositeFormat and
             the connected magic, yet
        """
        new = analysis_acc is None
        if new:
            analysis_acc = dict(filter_backtrack={},
                                terminal_chain=[[]])
        filter_backtrack = analysis_acc['filter_backtrack']
        terminal_chain = analysis_acc['terminal_chain'][-1]

        assert tuplist(filter_chain) and filter_chain
        # PASSDOWN or FILTERS
        passed_filter_length = len(filter_chain)
        pass_through, filter_chain = head_tail(*filter_chain) \
                                     if len(filter_chain) > 1 \
                                     and (not isinstance(filter_chain[0], tuple) \
                                     or len(filter_chain[0]) < 2) \
                                     else (None, filter_chain)
        # the condition handles differences between:
        #    ('A',
        #        ('B'),
        #        ('C'))
        # and
        #    ('A',
        #        ('B',
        #            ('C')))
        # note also need to treat correctly, e.g.:
        #    (
        #      ('A',
        #          ('B')),
        #      ('C',
        #          ('B'))
        #    )
        # XXX: regardless if isinstance(filter_chain[1], tuple)
        if len(filter_chain) >= passed_filter_length + int(new):
            filter_chain = (filter_chain, )
        for i_origin in filter_chain:
            if not i_origin:
                continue
            i, i_tail = head_tail(i_origin)
            # bt denotes filters feeding this one
            bt = filter_backtrack.setdefault(i, OrderedDict())
            if new or not (bt or i_tail):  # preorder
                # new for UPFILTERs, which are also terminals (input ones)
                if new and terminal_chain:
                    if isinstance(terminal_chain[0], list):
                        terminal_chain = terminal_chain.append([i])
                    else:
                        terminal_chain[:] = [terminal_chain[:]] + [[i]]
                else:
                    terminal_chain.append(i)
            if pass_through:
                if pass_through in bt:
                    raise CommandError(me,
                        "filter `{0}' is feeded by `{1}' more than once",
                        i.__class__.__name__, pass_through.__class__.__name__
                    )
                com = None  # for when CompositeFormat involved
                if (hasattr(pass_through.out_format, '_protocols')
                        and hasattr(i.in_format, '_protocols')):
                    com = pass_through.out_format.common_protocols(i.in_format)
                    if not com:
                        raise CommandError(me,
                            "filter `{0}' and its feeder `{1}' have no protocol"
                            " in common",
                            i.__class__.__name__, pass_through.__class__.__name__
                        )
                bt[pass_through] = com
            if i_tail:
                # PASSDOWN
                # this uses a dirty trick of exploiting the end of the list
                # as a sort of a stack, where the per-recursion-level result
                # is available for the caller (who is also responsible for
                # preparing a new list here for callee to fill) so it can
                # move it to the right position afterwards
                analysis_acc['terminal_chain'].append([])  # not terminal_chain
                # see "the condition handles differences between" comment
                me(i_origin, analysis_acc)
                # postorder
                ret = analysis_acc['terminal_chain'].pop()
                if ret:
                    # not a another use of already used (merging) filter
                    terminal_chain.append(ret)
            elif new:
                # yes, terminal UPFILTER is tracked twice as terminal (I/O)
                terminal_chain.append(i)

        return analysis_acc

    #
    # self-introspection (arguments, description, options)
    #

    @MimicMeta.method
    def _figure_parser_opt_dumpnoop(self, options, shortopts):
        choices = []
        for fname, f in iter_items(self._filters):
            if issubclass(f.in_format.__class__, f.out_format.__class__):
                choices.append(fname)
        # XXX NOOPizing doesn't make sense for input filters?
        debug_opts = (
            ('noop', False,
             "debug only: NOOPize filter (2+: repeat) [none out of %choices]"),
            ('dump', True,
             "debug only: dump (intermediate) output of the filter (2+: repeat)"
             " [none out of %choices]"),
        )
        for optname_used, universal, help_text in debug_opts:
            short_aliases = shortopts.setdefault(optname_used[0], [])
            assert optname_used not in \
                (options[i][0][0] for i in short_aliases)
            log.debug("choices: {0}".format(choices))
            opt = dict(
                action='append',
                choices=choices + ['ANY'] if universal else choices,
                default=[],
                expert=True,
                help=help_text,
            )
            options.append([["--" + optname_used], opt])

    @MimicMeta.method
    def _figure_parser_opt_unofficial(self, options, shortopts, fnc_varnames):
        # unofficial/unsupported ones  (XXX shortopts unused)
        for var in fnc_varnames:
            optname_used = cli_decor(var)
            options.append([["--" + optname_used], dict(
                expert=True,
                help="(undocumented expert option)",
            )])

    @MimicMeta.method
    def _figure_parser_desc_opts(self, fnc_defaults, fnc_varnames,
                                 opt_group=None):
        readopts, common_tail = False, False
        shortopts, options, expert =  OrderedDict(), [], []
        description = []
        fnc_varnames = set(fnc_varnames)
        opt_group = opt_group or OptionParser()

        for line in self.__doc__.splitlines():
            line = line.lstrip()
            if readopts:
                if not line:
                    continue
                if line.lower().startswith(CMD_HELP_OPTSEP_COMMON):
                    common_tail = True
                    continue
                line = line.replace('\t', ' ')
                optname, optdesc = head_tail(line.split(' ', 1))  # 2nd->tuple
                if not all((optname, optdesc)) or optname not in fnc_varnames:
                    log.warning("Bad option line: {0}".format(line))
                else:
                    target = expert if optname.startswith('_') else options
                    optname_used = cli_decor(optname.lstrip('_'))
                    log.debug("Command `{0}', found option `{1}' ({2})".format(
                        self.__class__.name, optname_used, optname
                    ))
                    fnc_varnames.remove(optname)
                    short_aliases = shortopts.setdefault(optname_used[0], [])
                    opt = {}
                    if target is expert:
                        opt['expert'] = True
                        opt['dest'] = optname  # (un)decor just works, '_' not
                    elif not common_tail:
                        assert optname_used not in \
                            (options[i][0][0] for i in short_aliases)
                        short_aliases.append(len(options))  # as an index
                    opt['help'] = optdesc[0].strip()
                    if optname in fnc_defaults:  # default if known
                        default = fnc_defaults[optname]
                        if default in (True, False):
                            opt['action'] = ('store_true',
                                             'store_false')[int(default)]
                            opt['help'] += " [{0}]".format('enabled' if default
                                                           else 'disabled')
                        else:
                            opt['help'] += " [%default]"
                        opt['default'] = default
                    target.append([["--" + optname_used], opt])
            elif line.lower().startswith(CMD_HELP_OPTSEP_PRIMARY):
                readopts = True
            else:
                description.append(line)

        for short in tuple(shortopts):  # foreach in ideal shorts, ditto in...
            for i, alias in enumerate(shortopts[short]):  # ... conflicting ones
                for c in longopt_letters_reprio(options[alias][0][0]):
                    use = '-' + c
                    if opt_group.has_option(use) or (i > 0 and c in shortopts):
                        continue
                    if c not in shortopts:
                        shortopts[c] = (i, )
                    break
                else:
                    log.info("Could not find short option for `{0}'"
                             .format(options[alias][0]))
                    break
                options[alias][0].append(use)

        self._figure_parser_opt_dumpnoop(options, shortopts)
        options.extend(expert)
        self._figure_parser_opt_unofficial(options, shortopts, fnc_varnames)

        description = description[:-1] if not description[-1] else description
        description = '\n'.join(description)
        return description, options

    @MimicMeta.method
    def parser_desc_opts(self, opt_group=None):
        """Parse docstring as description + Option constructor args list"""
        if self._desc_opts is None:
            self._desc_opts = self._figure_parser_desc_opts(
                *self._fnc_defaults_varnames, opt_group=opt_group
            )
        return self._desc_opts

    #
    # execution related
    #

    @MimicMeta.staticmethod
    @selfaware
    def _iochain_check_terminals(me, io_chain, terminal_chain, magic_fds,
                                 interpolations={}):
        # validate "terminal filter chain" vs "io chain" while solving magic_fds
        # 1. "shapes" match incl. input (head)/output (tail) protocol match
        if len(terminal_chain) == 1 and len(io_chain) == len(terminal_chain[0]):
            # see `deco`: 2.
            io_chain = args2tuple(io_chain)
        ret = []
        to_check = apply_loose_zip_preserving_depth(terminal_chain, io_chain)
        if to_check and fltiodecl(to_check[0]):
            to_check = [to_check]  # restore _iochain_proceed-clipped wrapping
        for to_check_inner in to_check:
            ret_to_check_inner = []
            for passno, check in enumerate(head_tail(to_check_inner)):
                checked = apply_aggregation_preserving_depth(
                    lambda i:
                        head_tail(protodictval(i[1]))[0] not in getattr(i[0],
                            ('in_format', 'out_format')[passno])._protocols
                            and str(head_tail(i[1])[0]) or None
                        if fltiodecl(i) else i if any(i) else None
                )(check)
                checked_flat = apply_intercalate((checked,))
                for order, proto in filter_u(lambda i, x: x,
                                             enumerate(checked_flat)):
                    if proto is zip_empty:
                        continue
                    raise CommandError(me,
                        "filter resolution #{0} of {1}: {2}", order + 1,
                        ('input', 'output')[passno],
                        "`{0}' filter/io chain definition (shape) mismatch"
                        .format(proto)
                        if isinstance(proto, (type(zip_empty), Filter))
                        else "`{0}' protocol not suitable".format(proto)
                    )
                # handle "magic files"
                resolved = apply_aggregation_preserving_depth(
                    lambda i:
                        (i[0], SimpleFormat.io_decl_specials(i[1], passno == 0,
                                                             magic_fds,
                                                             interpolations))
                        if fltiodecl(i) else i
                )(check)
                ret_to_check_inner.append(resolved)
            ret.append(ret_to_check_inner)
        return ret

    @MimicMeta.method  # should be classmethod?
    def _iochain_proceed(self, cmd_ctxt, io_chain):
        # currently works sequentially, jumping through the terminals in-order;
        # when any of them (apparently the output one) hasn't its prerequisites
        # (i.e., input data) satisfied, the run is restarted with first
        # producing such data (which are output of another filter feeding
        # the one in question) -- this can be repeated multiple times if
        # there is a longer chain forming such a gap
        # -- this is certainly needlessly slow method, but there is a hope
        #    the same approach could be applied when parallelizing the stuff
        # XXX could be made more robust (ordering still not as strict as it
        #                                should)
        # XXX some parts could be performed in parallel (requires previous
        #     item so to prevent deadlocks on cond. var. wait)
        #     - see also `heapq` standard module
        filter_backtrack = cmd_ctxt['filter_chain_analysis']['filter_backtrack']
        terminal_chain = cmd_ctxt['filter_chain_analysis']['terminal_chain'][-1]
        terminals = apply_intercalate(terminal_chain)

        native_fds = dict((f.fileno(), f) for f in (stderr, stdin, stdout))
        magic_fds = native_fds.copy()
        # XXX using with cmd_ctxt.prevented_taint() would be too pedantic
        #     and, furthermore, would destroy order-preserving because of
        #     OrderedDict to plain dict artificial "downcasting"
        terminal_chain = self._iochain_check_terminals(io_chain,
                                                       terminal_chain,
                                                       magic_fds,
                                                       cmd_ctxt['__filters__'])

        input_cache = cmd_ctxt.setdefault('input_cache', {}, bypass=True)
        worklist = list(reversed(tailshake(terminal_chain,
                                           partitioner=lambda x:
                                           not (tuplist(x)) or fltiodecl(x))))
        # if any "EMPTY" (zip_empty) value present, respective class name ~ str
        unused, tstmp = {}, hex(int(time()))[2:]
        while worklist:
            workitem = worklist.pop()
            if workitem == zip_empty:
                log.debug("Worklist: EMPTY value observed, skipped")
                continue
            flt, io_decl = workitem
            io_decl_use = protodictval(io_decl)
            io_decl, passout = (io_decl_use, unused if io_decl_use is io_decl
                                             else io_decl)
            flt_ctxt = cmd_ctxt.ensure_filter(flt)
            with flt_ctxt.prevented_taint():
                fmt_kws = filterdict_keep(flt_ctxt, *flt.in_format.context)
            if not filter_backtrack[flt] and 'out' not in flt_ctxt:
                # UPFILTER in in-mode
                log.debug("Run `{0}' filter with `{1}' io decl. as UPFILTER"
                          .format(flt.__class__.__name__, io_decl))
                if io_decl in input_cache:
                    in_obj = input_cache[io_decl]
                else:
                    with cmd_ctxt.prevented_taint():
                        in_obj = flt.in_format.as_instance(*io_decl, **fmt_kws)
                    input_cache[io_decl] = flt_ctxt['in'] = in_obj
            elif filter_backtrack[flt] and 'out' not in flt_ctxt:
                # not UPFILTER in either mode (nor output already precomputed?)
                log.debug("Run `{0}' filter with `{1}' io decl. as DOWNFILTER"
                          .format(flt.__class__.__name__, io_decl))
                ok, notyet = bifilter(lambda x: 'out' in
                                        cmd_ctxt.filter(x.__class__.__name__),
                                      filter_backtrack[flt])
                if notyet:
                    log.debug("Backtrack with inclusion of {0} to feed `{1}'"
                              .format(', '.join("`{0}'"
                                      .format(ny.__class__.__name__)
                                              for ny in notyet),
                                      flt.__class__.__name__))
                    worklist.append((flt, io_decl if passout is unused
                                          else passout))
                    worklist.extend(reversed(tuple((ny, None)
                                             for ny in notyet)))
                    continue

                inputs = tuple(cmd_ctxt.filter(x.__class__.__name__).get('out')
                               for x in filter_backtrack[flt])
                assert all(inputs)
                with cmd_ctxt.prevented_taint():
                    in_obj = flt.in_format.as_instance(*inputs, **fmt_kws)
                flt_ctxt['in'] = in_obj  # referred in interpolation -> a bug?
            if 'out' not in flt_ctxt or flt not in terminals:
                if 'out' not in flt_ctxt:
                    if flt.__class__.name in cmd_ctxt['filter_noop']:
                        ret = in_obj
                    else:
                        # re io_decl: allow terminal filters have a peek at
                        # respective resolved(!) filter IO declaration, so they
                        # can, e.g., choose a final formatting (see cmd-wrap)
                        # XXX useful just for output terminals, really
                        if flt in terminals:
                            flt_ctxt['io_decl'] = io_decl
                        with cmd_ctxt.prevented_taint():
                            ret = flt(in_obj, flt_ctxt)
                        if flt in terminals:
                            flt_ctxt.pop('io_decl')
                    flt_ctxt['out'] = ret
                if flt not in terminals or not filter_backtrack[flt]:
                    if (flt.__class__.name in cmd_ctxt['filter_dump']
                        or 'ANY' in cmd_ctxt['filter_dump']):
                        try:
                            fn = 'dump-{0}-{1}-{2}'.format(
                                flt.__class__.name,
                                flt_ctxt['in'].hash,
                                tstmp,
                            )
                            ret(SimpleFormat.FILE, fn)
                        except FormatError:
                            flt_ctxt.ctxt_svc_output("dumping failed",
                                                     base='error', urgent=True)
                        else:
                            flt_ctxt.ctxt_svc_output("|subheader:dump:|"
                                                     " |highlight:{0}|"
                                                     .format(fn))
                    continue
            # output time!  (incl. UPFILTER terminal listed twice in io_chain)
            with cmd_ctxt.prevented_taint():  # still needed for late binding
                io_decl = SimpleFormat.io_decl_specials(io_decl, 0, magic_fds,
                                                        cmd_ctxt['__filters__'])
            log.debug("Run `{0}' filter with `{1}' io decl. as TERMINAL"
                      .format(flt.__class__.name, io_decl))
            # store output somewhere, which even can be useful (use as a lib)
            passout['passout'] = flt_ctxt['out'](*io_decl)
            if passout is unused and io_decl[0] == SimpleFormat.FILE:
                flt_ctxt.ctxt_svc_output("|subheader:output:| |highlight:{0}|"
                                         .format(passout['passout']))

        # close "magic" fds
        foreach(lambda k: k in native_fds or magic_fds[k].close(), magic_fds)
        return EC.EXIT_SUCCESS  # XXX some better decision?

    @MimicMeta.method
    def __call__(self, opts, args=None, cmd_ctxt=None):
        """Proceed the command"""
        ec = EC.EXIT_SUCCESS
        maxl = len(sorted(self._filters, key=len)[-1])
        color = dict(auto=None, never=False, always=True)[
            getattr(opts, 'color', 'auto')
        ]
        cmd_ctxt = cmd_ctxt or CommandContext({
            'filter_noop':           getattr(opts, 'noop', ()),
            'filter_dump':           getattr(opts, 'dump', ()),
            'system':                getattr(opts, 'sys', ''),
            'system_extra':          tuple(se for se in
                                           getattr(opts, 'dist', '').split(',')
                                           if se),
            'svc_output':            FancyOutput(f=stderr,
                                                 quiet=getattr(opts, 'quiet',
                                                               False),
                                                 prefix=("|header:[{{0:{0}}}]| "
                                                         .format(maxl)),
                                                 color=color,
                                     ),
            'color':                color,
        }, bypass=True)
        cmd_ctxt.setdefault('filter_chain_analysis',
                            self.filter_chain_analysis, bypass=True)
        cmd_ctxt.ensure_filters(iter_values(self._filters))
        kwargs = {}
        # desugaring, which is useful mainly if non-contiguous sequence
        # of value-based options need to be specified
        args = [None if not a else a for a in args[0].split('::')] + args[1:] \
               if args else []
        args.reverse()  # we will be poping from the end
        for v in self._fnc_defaults_varnames[1]:
            default = self._fnc_defaults_raw.get(v, None)
            opt = getattr(opts, v, default)
            if isinstance(opt, basestring):
                try:
                    opt = opt.format(**cmd_ctxt['__filters__'])
                    # XXX type adjustment at least for bool?
                except (AttributeError, ValueError, KeyError):
                    # AttributeError ~ may be available later on,
                    #                  resolved in io_decl_specials
                    pass
            if isinstance(opt, MutableMapping) \
                    or not isinstance(default, basestring) \
                        and isinstance(opt, basestring) \
                    or areinstancesupto(opt, default, object, type) \
                        and opt != default:
                kwargs[v] = opt
                continue
            elif not isinstance(opt, (basestring, type(None))):
                log.info("`{0}' command: skipping attempt to pair argument"
                         " to non-string `{1}' option (specify whole option"
                         " instead)".format(self.__class__.name, v))
                continue
            try:
                cur = args.pop()
                while cur == '':  # deliberately skip (implicit) empty string
                    cur = args.pop()
                if cur is not None:  # incl. case of explicit empty string
                    kwargs[v] = cur.replace("''", "") if len(cur) == 2 else cur
                    continue
                raise IndexError  # "required arg not provided" for sugar spec
            except IndexError:
                if opt is not None:
                    continue
                raise CommandError(self, "missing ex-/implicit `{0}' value", v)
        if args:
            log.info("`{0}' command: unconsumed arguments: {1}"
                     .format(self.__class__.name, ', '.join("`" + a + "'"
                                                            for a in args)))
        log.debug("Running command `{0}';  args={1}, kwargs={2}"
                  .format(self.__class__.name, args, kwargs))
        io_driver = any2iter(self._fnc(cmd_ctxt, **kwargs))
        io_handler = (self._iochain_proceed, lambda c, ec=EC.EXIT_SUCCESS: ec)
        io_driver_map = zip_longest(io_driver, io_handler)
        for driver, handler in io_driver_map:
            driver = () if driver is None else (driver, )
            ec = handler(cmd_ctxt, *driver)
            if ec != EC.EXIT_SUCCESS:
                break
        return ec

    @MimicMeta.classmethod
    def deco(cls, *filter_chain):
        """Decorator as an easy factory of actual commands

        Parameters:
            filter_chain: particular scalars and vectors (variable depth)
                          representing graph of filters that form this command

        Note on graph representation within filter_chain:

                 __B    ___D
                /      /
            A--<___C--<             in ----------------> out
                       \
            O___________>--P


        graph with letter denoting the filters and with the left->right
        direction of flow from the inputs towards outputs (final outputs
        at terminals: B, D, P), is encoded as:

            (A, B, (C, D, P)), (O, P)

        when reformatted as per the only Approved Indenting Convention (TM):

            (
               (A,
                   (B),
                   (C,
                       (D),
                       (P))),
               (O,
                       (P))
            )

        where, for filter x (in {A, ..., D, O, P} for the example at hand):

            FILTERCHAIN ::= UPFILTERS
            UPFILTERS   ::= TERMINAL | ( FILTERS )
            FILTERS     ::= FILTER, | FILTERS FILTER
            FILTER      ::= PASSDOWN | TERMINAL
            PASSDOWN    ::= (TERMINAL, FILTERS)  # ~ DOWNFILTERS from that down
            TERMINAL    ::= x

        where:
            - {UP,DOWN}FILTERS dichotomy is present only as
              a forward-reference for easier explanation;
              it's asymmetric, meaning that UPFILTERS are terminals
              in the graph, whereas DOWNFILTERS can also be PASSDOWNs
            - there is a limitation such that each filter can
              be contained as at most one node in the graph as above
              (this corresponds to the notion of inputs merge for
              the filter, as otherwise there would be ambiguity:
              in the selected notation, can the filter occurences stand
              for unique nodes?  remember, filters as singletons)
            - UPFILTERS ::= TERMINAL is a syntactic sugar exploiting
              unambiguity in converting such expression as (TERMINAL, )
            - to make it explicit, the graph is expressed in depth-first
              (or DFS) manner


        Note on the decorated function:
            It should either return an iterable or behave itself as a generator
            yielding the items (at once) and on subsequent round triggering
            some postprocessing (still from decorated function's perspective).
            The items coming from the function encodes the protocols at
            the input(s) and the output(s) of the filter graph encoded in
            `filter_chain` and ought to reflect this processing construct
            as follows:
                1. for each UPFILTER in order, there is a tuple of two parts
                   1b. first part denotes the input (only single decl)
                   2b. second part denotes the output, which follows the
                       branch of filter chain pertaining the particular
                       UPFILTER, and can be either scalar or (arbitrarily)
                       nested iterable to match that filter chain branch
                       (proper nesting is not needed, only the order is
                       important, see point 4.)
                2. if there is just one UPFILTER, the toplevel definition
                   can be just the respective un-nested item, as this case
                   is easy to distinguish and apply un-sugaring if applicable
                3. when there is the same filter down the line shared by
                   2+ UPFILTERs (cf. "limitation such that each filter" above)
                   the respective protocol encoding is expected just once
                   within the first(!) respective UPFILTER definition
                #-- not yet, if ever, as it is opposed by good diagnostics --
                #4. nesting of the second part of the tuple (2b.) is not
                #   strictly needed and only the order is important,
                #   as the association is performed on intercalated chains
                #   anyway (note that this is orthogonal to simplification 2.)

            for the graph above, it would be -- following the only
            Approved Indenting Convention (TM) --, e.g.,:

                (
                    ('Aproto', 'a-in.txt'),
                    (
                        ('Bproto', 'b-out.txt'),
                        (
                            ('Dproto', 'd-out.txt'),
                            ('Pproto'),
                        ),
                    ),
                ),
                (
                    ('Oproto', 'o-in.txt'),
                )

            #which, as per point 4., can be further simplified as:

            #(('Aproto', 'a-in.txt'),
            #    ('Bproto', 'b-out.txt'), ('Dproto', 'd-out.txt'), ('Pproto')),
            #(('Oproto', 'e-in.txt'), )
        """
        def deco_fnc(fnc):
            log.debug("Command: deco for {0}"
                      .format(fnc))
            fnc_defaults, fnc_varnames, wrapped = defer_common(fnc, skip=1)
            attrs = {
                '__module__': fnc.__module__,  # original
                '__doc__': wrapped.__doc__,
                '_filter_chain': filter_chain,
                '_fnc': staticmethod(wrapped),
                '_fnc_defaults_varnames': (fnc_defaults, fnc_varnames),
                '_fnc_defaults_raw': fnc_defaults.copy(),  # un-interpolated
            }
            # optimization: shorten type() -> new() -> probe
            ret = cls.probe(fnc.__name__, (cls, ), attrs)
            return ret
        return deco_fnc


Command = MimicMeta('Command', commands, _Command)


class _CommandAlias(object):
    """Way to define either static or dynamic command alias"""

    @MimicMeta.method
    def __new__(cls, flts, cmds, *args):
        ic, sys, sys_extra = (lambda i={}, s='', e='', *a: (i, s, e))(*args)
        # XXX really pass mutable cmds dict?
        use_obj = cls
        use_obj = use_obj._fnc(cmds, sys,  # see Command.__call__ (cmd_ctxt)
                               tuple(se for se in sys_extra.split(',') if se))
        for i in xrange(1, 100):  # prevent infloop by force
            if isinstance(use_obj, basestring):
                use_obj = cmds.get(use_obj, None)
                if not isinstance(use_obj, (nonetype, Command)):
                    assert issubclass(use_obj, CommandAlias)
                    assert use_obj is not cls, "trivial infloop"
                    continue
            elif use_obj is None:
                pass
            else:
                assert issubclass(use_obj, (Command, CommandAlias))
                if use_obj in ic and ic[use_obj] in cmds:
                    use_obj = cmds[ic[use_obj]]
                else:
                    if use_obj in ic:
                        log.warning("Resolve at `{0}' command: already proved"
                                    " unresolvable(?)".format(use_obj.name))
                    name = '_' + use_obj.name
                    assert name not in cmds
                    ic[use_obj] = name
                    cmds[name] = use_obj = use_obj(flts, cmds, *args)
            assert isinstance(use_obj, (nonetype, Command)), repr(use_obj)
            return use_obj

    @MimicMeta.classmethod
    def deco(outer_cls, decl):
        if not hasattr(decl, '__call__'):
            assert issubclass(decl, Command)
            fnc = lambda *args, **kwargs: decl
        else:
            fnc = decl
        log.debug("CommandAlias: deco for {0}".format(fnc))

        attrs = dict(
            __module__=fnc.__module__,
            _fnc=staticmethod(fnc)
        )
        # optimization: shorten type() -> new() -> probe
        ret = outer_cls.probe(fnc.__name__, (outer_cls, ), attrs)
        return ret


CommandAlias = MimicMeta('CommandAlias', commands, _CommandAlias)



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""Format representing merged/isolated (1/2 levels) of single command to exec"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
from logging import getLogger
from shlex import split

log = getLogger(__name__)

from ..format import SimpleFormat
from ..protocol import Protocol
from ..utils import head_tail
from ..utils_2to3 import bytes_enc, iter_items
from ..utils_func import add_item, apply_intercalate, apply_partition

# man bash | grep -A2 '\s\+metacharacter$'
_META_CHARACTERS =  '|', '&', ';', '(', ')', '<', '>' , ' ', '\t'
# man bash | grep -A2 '\s\+control operator$' (minus <newline>)
_CONTROL_OPERATORS = '||', '&', '&&', ';', ';;', '(', ')', '|', '|&'  # , '\n'
# man bash | grep -A4 '^RESERVED WORDS$' | tail -n1
_RESERVED_WORDS = ('!', 'case', 'coproc', 'do', 'done', 'elif', 'else', 'esac',
                   'fi', 'for', 'function', 'if', 'in', 'select', 'then',
                   'until', 'while', '{', '}', 'time', '[[', ']]')

_META_WORDS = _CONTROL_OPERATORS + _RESERVED_WORDS

ismetaword = \
    lambda x: \
    x in _META_WORDS \
        or x.endswith('(') and x[-2:-1] in '$<' \
            and x[:-2].rstrip(' \t') in ('', '"') \
        or x.startswith(')') and x[1:].lstrip(' \t') in ('', '"')


class command(SimpleFormat):
    native_protocol = SEPARATED = Protocol('separated')
    BYTESTRING = SimpleFormat.BYTESTRING
    DICT = Protocol('dict')
    MERGED = Protocol('merged')

    @staticmethod
    def _escape(base, qs=("'", '"')):
        # rule: last but one item in qs cannot be escaped inside enquotion
        # XXX: shell constructs like "&&" and "<>" need to be blank-separated
        #      and still there may be a confusion if they are used alone
        #      (fortunately, pcs uses switches and "X=Y" syntax most of
        #      the time); real solution would be to implement complete
        #      context-free(?) grammar of the POSIX shell, which is currently
        #      a blatant overkill
        ret = []
        for b in base:
            if (b.strip(''.join(_META_CHARACTERS)) and
                    any(c in b for c in qs + tuple('#$') + _META_CHARACTERS)):
                use_q, prologue, epilogue = ('', ) * 3
                if not(b.endswith('(') and b[-2:-1] in '$<'
                       and b[:-2].rstrip(' \t') in ('', '"')
                       or
                       b.startswith(')') and b[1:].lstrip(' \t') in ('', '"')):
                    if b.startswith('<<<'):
                        _, prologue, b = b.partition('<<<')
                    use_qs = tuple(c for c in qs if c != "'") if '$' in b else qs
                    for q in use_qs:
                        if q not in b:
                            use_q = q
                            break
                    else:
                        use_q = use_qs[-1]
                        if use_q != qs[0]:
                            b = b.replace(use_q, '\\' + use_q)
                        else:
                            raise RuntimeError('cannot quote the argument')
                b = b.join((use_q, use_q)).join((prologue, epilogue))
            ret.append(b)
        return ret

    @SimpleFormat.producing(BYTESTRING, chained=True)
    def get_bytestring(self, *iodecl):
        """Return command as canonical single string"""
        # chained fallback
        return bytes_enc(' '.join(self.MERGED(protect_safe=True)))

    @SimpleFormat.producing(SEPARATED, protect=True)
    def get_separated(self, *iodecl):
        merged = self.MERGED()
        merged.reverse()
        ret, acc, takes = [], [], 2  # by convention, option takes at most 1 arg
        while merged:
            i = merged.pop()
            # treat `-OPT` followed with `-NUMBER` just as if it was `NUMBER`,
            # also add an extra split on "meta words"
            if acc == ['--'] or i is None or ismetaword(i) \
                    or i.startswith('-') and i != '-' and not i[1:].isdigit():
                if not acc:
                    pass
                elif acc[0].startswith('-'):
                    ret.extend(filter(bool, (tuple(acc[:takes]),
                                             tuple(acc[takes:]))))
                else:
                    ret.append(tuple(acc))
                acc, takes = [] if i is None else [i], 2  # reset option-arg cnt
            elif self._dict.get('magic_split', False):
                split = i.split('::')  # magic "::"-split
                if len(acc) == 1 and acc[0].startswith('-') and acc[0] != '-':
                    takes = len(split) + 1  # sticky option multi-arguments
                acc.extend(split)
            else:
                acc.append(i)
            if acc and not merged:
                merged.append(None)  # mark terminal acc -> ret propagation
        return ret

    @SimpleFormat.producing(MERGED, protect=True)
    def get_merged(self, *iodecl):
        if self.BYTESTRING in self._representations:  # break the possible loop
            # XXX backticks not supported (yet)
            preprocessed = apply_partition(
                self.BYTESTRING(),
                lambda x, _, acc:
                    x == '(' and acc and not(acc[0].lstrip().startswith('#'))
                             and acc[-1][-1:] in '$<'
                             and not(any((
                                (lambda y, z: 0 if z == 1 and y == '"' else z % 2)(
                                    y, acc[-1].lstrip(' \t' + y).count(y) - acc[-1].count('\\' + y)
                                )
                             ) for y in "'\""))
                    or
                    x == ')' and acc and (not acc[-1] or acc[-1][-1] != '\\')
                             and any(y in acc[-2:-1] for y in '()')
                             and not(any((
                                acc[-1].lstrip(' \t' + y).count(y) - acc[-1].count('\\' + y)
                             ) % 2 for y in "'\""))
                    or
                    x == ';' and acc and not(acc[0].lstrip().startswith('#'))
                             and not(any((
                                acc[-1].lstrip(' \t' + y).count(y) - acc[-1].count('\\' + y)
                             ) % 2 for y in "'\""))
            )
            ret, prev = [], ''
            # XXX could check quotation pairs match using a stack
            for p in add_item(preprocessed, ''):
                tail1, tail2 = '', ''
                if p == '(':
                    if p == '(' and prev[-1:] in '$<':
                        qs = tuple(
                            x for x in '"' if
                            (prev[:-1].count(x) - prev[:-1].count('\\' + x)) % 2
                        )
                        if qs and prev[:-1].rstrip(" \t")[-1:] in qs:
                            prev, tail1, tail2 = prev.rpartition(qs[0])
                    ret.extend(filter(bool, split(prev) + [tail1 + tail2]))
                    ret[-1] += p
                    p = ''
                elif prev == ')':
                    qs = tuple(
                        x for x in '"' if
                        (p[:-1].count(x) - p[:-1].count('\\' + x)) % 2
                    )
                    if qs and p.lstrip(" \t")[:1] in qs:
                        tail1, tail2, p = p.partition(qs[0])
                    ret.extend(filter(bool, [prev + tail1 + tail2]))
                elif prev:
                    ret.extend(split(prev))
                prev = p
            if self._dict.get('enquote', True):
                ret = self._escape(ret)
            offset = 0
            for i, lexeme in enumerate(ret[:]):
                # heuristic(!) method to normalize: '-a=b' -> '-a', 'b'
                if (lexeme.count('=') == 1 and lexeme.startswith('-') and
                    ('"' not in lexeme or lexeme.count('"') % 2) and
                    ("'" not in lexeme or lexeme.count("'") % 2)):
                    ret[i + offset:i + offset + 1] = lexeme.split('=')
                    offset += 1
        elif self.DICT in self._representations:  # break the possible loop (2)
            d = self.DICT(protect_safe=True)
            if not isinstance(d, OrderedDict):
                log.warning("'{0}' format: not backed by OrderedDict".format(
                    self.__class__.name
                ))
            ret = list(d.get('__cmd__', ()))
            ret.extend((k, v) for k, vs in iter_items(d) for v in (vs or ((), ))
                                        if k not in ('__cmd__', '__args__'))
            ret.extend(d.get('__args__', ()))
        else:
            ret = self.SEPARATED(protect_safe=True)
        return apply_intercalate(ret)

    @SimpleFormat.producing(DICT, protect=True)
    # not a perfectly bijective mapping, this is a bit lossy representation,
    # on the other hand it canonicalizes the notation when turned to other forms
    def get_dict(self, *iodecl):
        separated = self.SEPARATED()
        separated.reverse()
        ret = OrderedDict()
        arg_bucket = '__cmd__'
        while separated:
            head, tail = head_tail(separated.pop())
            if head.startswith('-') and head != '-':
                arg_bucket = '__args__'
            else:
                head, tail = arg_bucket, head
            ret.setdefault(head, []).append(tail)
        return ret



# -*- coding: UTF-8 -*-
# Copyright 2016 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ....filters._2pcscmd import verbose_ec_test, verbose_inform
from ....utils_xslt import NL, xslt_is_member

cib2pcscmd_options = (
    'symmetrical',
    'score',
    'kind',
)

cib2pcscmd_set_options = (
    'action',
    'sequential',
    'require-all',
    # following seems not to be reflected in the schemas at all
    #'kind',
)

cib2pcscmd = ('''\
    <xsl:choose>
        <!-- plain first-then -->
        <xsl:when test="@first and @then">
''' + (
            verbose_inform('"new order constraint: ", @id')
) + '''
            <xsl:value-of select="concat($pcscmd_pcs, 'constraint order')"/>
            <xsl:if test="@first-action and @first-action != 'start'">
                <xsl:value-of select="concat(' ', @first-action)"/>
            </xsl:if>
            <xsl:value-of select="concat(' ', @first, ' then')"/>
            <xsl:if test="@then-action and @then-action != 'start'">
                <xsl:value-of select="concat(' ', @then-action)"/>
            </xsl:if>
            <xsl:value-of select="concat(' ', @then)"/>
            <xsl:for-each select="@*[
''' + (
                xslt_is_member('name()', cib2pcscmd_options)
) + ''']">
                <xsl:value-of select="concat(' ', name(), '=', .)"/>
            </xsl:for-each>
            <xsl:value-of select="concat(' ', 'id=', @id)"/>
            <xsl:value-of select="'%(NL)s'"/>
''' + (
            verbose_ec_test
) + '''
        </xsl:when>

        <!-- resource sets -->
        <xsl:when test="resource_set">
''' + (
            verbose_inform('"new order constraint (resource set): ", @id')
) + '''
            <xsl:value-of select="concat($pcscmd_pcs, 'constraint order')"/>

            <xsl:for-each select="resource_set">
                <xsl:value-of select="' set'"/>
                <xsl:for-each select="resource_ref">
                    <xsl:value-of select="concat(' ', @id)"/>
                </xsl:for-each>

                <xsl:for-each select="@*[
''' + (
                    xslt_is_member('name()', cib2pcscmd_set_options)
) + ''']">
                    <xsl:value-of select="concat(' ', name(), '=', .)"/>
                </xsl:for-each>

            </xsl:for-each>

            <xsl:value-of select="' setoptions'"/>
            <xsl:for-each select="@*[
''' + (
                xslt_is_member('name()', cib2pcscmd_options)
) + ''']">
                <xsl:value-of select="concat(' ', name(), '=', .)"/>
            </xsl:for-each>
            <xsl:value-of select="concat(' ', 'id=', @id)"/>
            <xsl:value-of select="'%(NL)s'"/>
''' + (
            verbose_ec_test
) + '''
        </xsl:when>
    </xsl:choose>

''') % dict(
    NL=NL,
)



# -*- coding: UTF-8 -*-
# Copyright 2014 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)

ccs_obfuscate_identifiers = '''\

    <!-- REL-FENCE-METHOD -->

    <xsl:copy>
        <xsl:apply-templates select="@*"/>
        <xsl:for-each select="node()">
            <xsl:choose>
                <xsl:when test="name() = 'method'">
                    <xsl:variable name="MethodPos"
                                  select="count(preceding-sibling::method) + 1"/>
                    <xsl:copy>
                        <xsl:copy-of select="@*"/>
                        <xsl:for-each select="@name">
                            <xsl:attribute name="{name()}">
                                <xsl:value-of select="concat(
                                    'REL-FENCE-METHOD-', $MethodPos
                                )"/>
                            </xsl:attribute>
                        </xsl:for-each>
                        <xsl:apply-templates/>
                    </xsl:copy>
                </xsl:when>
                <xsl:otherwise>
                    <xsl:copy>
                        <xsl:apply-templates/>
                    </xsl:copy>
                </xsl:otherwise>
            </xsl:choose>
        </xsl:for-each>
    </xsl:copy>
'''



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ....filters._2pcscmd import verbose_ec_test, verbose_inform
from ....filters.cib2pcscmd import attrset_xsl
from ....utils_xslt import NL, xslt_is_member

cib2pcscmd_globaltags = (
    'network',
    'storage',
    'primitive',
)

cib2pcscmd_networkattrs = (
    'ip-range-start',
    'control-port',
    'host-interface',
    'host-netmask',
)

cib2pcscmd_portmapattrs = (
    'id',
    'port',
    'internal-port',
    'range',
)

cib2pcscmd_stmapattrs = (
    'id',
    'source-dir',
    'source-dir-root',
    'target-dir',
    'options',
)

cib2pcscmd = ('''\
    <xsl:choose>
        <xsl:when test="$pcscmd_extra_bundle">
''' + (
            verbose_inform('"new bundle: ", @id')
) + '''
            <xsl:value-of select="concat($pcscmd_pcs, 'resource bundle create ', @id,
                                         ' container ')"/>
            <xsl:for-each select="*[not(
''' + (
                xslt_is_member('name()', cib2pcscmd_globaltags)
) + ''')]">
                <xsl:choose>
        
                    <!-- skip the alleged container type duplications -->
                    <xsl:when test="position() != 1"/>
        
                    <!-- the nested snippets should be guarded with
                         xsl:when test="name() = 'foo'" -->
                    <clufter:descent-mix at="*"/>
        
                    <xsl:otherwise>
                        <xsl:message terminate="yes">
                            <xsl:value-of select="concat('WARNING: container type `',
                                                         name(), '` is currently',
                                                         ' unhandled')"/>
                        </xsl:message>
                    </xsl:otherwise>
        
                </xsl:choose>
            </xsl:for-each>
        
            <!-- network options -->
        
            <xsl:if test="network/@*[
 ''' + (
                xslt_is_member('name()', cib2pcscmd_networkattrs)
) + '''   
            ]">
                <xsl:value-of select="' network '"/>
            </xsl:if>
        
            <xsl:for-each select="network/@*">
                <xsl:choose>
                    <xsl:when test="
''' + (
                    xslt_is_member('name()', cib2pcscmd_networkattrs)
) + '''">
                        <xsl:value-of select='concat(" &apos;", name(), "=", ., "&apos;")'/>
                    </xsl:when>
                    <xsl:otherwise>
                        <xsl:message>
                            <xsl:value-of select="concat('WARNING: dropping non-whitelisted',
                                                         ' bundle/network option: `',
                                                         name(), '`')"/>
                        </xsl:message>
                    </xsl:otherwise>
                </xsl:choose>
            </xsl:for-each>
        
            <!-- port-mapping options (possibly multi-spec) -->
        
            <xsl:for-each select="network/port-mapping">
                <xsl:if test="@*[
 ''' + (
                    xslt_is_member('name()', cib2pcscmd_portmapattrs)
) + '''   
                ]">
                    <xsl:value-of select="' port-map '"/>
                </xsl:if>
        
                <xsl:for-each select="@*">
                    <xsl:choose>
                        <xsl:when test="
''' + (
                        xslt_is_member('name()', cib2pcscmd_portmapattrs)
) + '''">
                            <xsl:value-of select='concat(" &apos;", name(), "=", ., "&apos;")'/>
                        </xsl:when>
                        <xsl:otherwise>
                            <xsl:message>
                                <xsl:value-of select="concat('WARNING: dropping non-whitelisted',
                                                             ' bundle/port-mapping option: `',
                                                             name(), '`')"/>
                            </xsl:message>
                        </xsl:otherwise>
                    </xsl:choose>
                </xsl:for-each>
            </xsl:for-each>
        
            <!-- storage-mapping options (possibly multi-spec) -->
        
            <xsl:for-each select="storage/storage-mapping">
                <xsl:if test="@*[
 ''' + (
                    xslt_is_member('name()', cib2pcscmd_stmapattrs)
) + '''   
                ]">
                    <xsl:value-of select="' storage-map '"/>
                </xsl:if>
        
                <xsl:for-each select="@*">
                    <xsl:choose>
                        <xsl:when test="
''' + (
                            xslt_is_member('name()', cib2pcscmd_stmapattrs)
) + '''">
                            <xsl:value-of select='concat(" &apos;", name(), "=", ., "&apos;")'/>
                        </xsl:when>
                        <xsl:otherwise>
                            <xsl:message>
                                <xsl:value-of select="concat('WARNING: dropping non-whitelisted',
                                                             ' bundle/storage-mapping option: `',
                                                             name(), '`')"/>
                            </xsl:message>
                        </xsl:otherwise>
                    </xsl:choose>
                </xsl:for-each>
            </xsl:for-each>

            <!-- meta attrs -->
            <xsl:choose>
                <xsl:when test="$pcscmd_extra_bundle_meta">
''' + (
                    attrset_xsl("meta_attributes", cmd='" meta"')
) + '''
                </xsl:when>
                <xsl:otherwise>
                    <xsl:message>%(bundle_meta_msg)s</xsl:message>
                </xsl:otherwise>
            </xsl:choose>

            <xsl:value-of select="'%(NL)s'"/>
''' + (
            verbose_ec_test
) + '''
        </xsl:when>
        <xsl:otherwise>
            <xsl:message>%(bundle_msg)s</xsl:message>
        </xsl:otherwise>
    </xsl:choose>
''') % dict(
    NL=NL,
    bundle_msg="WARNING: target pacemaker/pcs versions do not support"
               " `bundle` construct, hence omitted",
    bundle_meta_msg="WARNING: target pacemaker/pcs versions do not support"
                    " meta attributes for `bundle` construct, hence omitted",
)



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)

###

# NOTE pcs doesn't support udpb directly via --transport (unless forced),
#      one has to specify --broadcast0 parameter instead and transport
#      will get silently switched to udpb

from ....utils_xslt import xslt_is_member

needlexml2pcscmd_transports = ('udp', 'udpu')
needlexml2pcscmd_rrp_mode = ('active', 'passive')
# 1:1 mapping of supported params to pcs arguments (not: rrpmode)
needlexml2pcscmd_supported = (
    'consensus',
    'join',
    'token',
    'token_coefficients',
    'fail_recv_const',
    'miss_count_const',
    'token_coefficient',
)

needlexml2pcscmd = '''\
    <!-- transport -->
    <xsl:if test="
''' + (
    xslt_is_member('@transport', needlexml2pcscmd_transports)
) + '''">
        <xsl:value-of select="concat(' --transport ', @transport)"/>
    </xsl:if>

    <!-- rrpmode -->
    <xsl:if test="
''' + (
    xslt_is_member('@rrp_mode', needlexml2pcscmd_rrp_mode)
) + '''">
        <xsl:value-of select="concat(' --rrpmode ', @rrp_mode)"/>
    </xsl:if>

    <clufter:descent-mix at="interface"/>

    <!-- ipv6? -->
    <xsl:if test="@ip_version = 'ipv6'">
        <xsl:value-of select="' --ipv6'"/>
    </xsl:if>

    <xsl:for-each select="@*[
''' + (
        xslt_is_member('name()', needlexml2pcscmd_supported)
) + ''']">
        <xsl:if test=".">
            <xsl:value-of select="concat(' --', name(), ' ', .)"/>
        </xsl:if>
    </xsl:for-each>

    <!-- corosync encryption, based on key/keyfile -->
    <xsl:choose>
        <xsl:when test="(
                          @key or @keyfile
                      ) and (
                          (
                            not(@secauth)
                            or
                            @secauth != 'off'
                          ) and (
                            not(@crypto_cipher)
                            or
                            @crypto_cipher != 'none'
                          )
                      )">
            <xsl:choose>
                <xsl:when test="$pcscmd_extra_corosync_encryption_forced
                                or
                                $pcscmd_extra_corosync_encryption">
                    <xsl:choose>
                        <xsl:when test="@keyfile and @keyfile != '/etc/corosync/authkey'">
                            <xsl:message>
                                <xsl:value-of select="concat('WARNING: `keyfile` granularity not',
                                                             ' supported, will at least enable',
                                                             ' encryption as such')"/>
                            </xsl:message>
                        </xsl:when>
                        <xsl:when test="@keyfile">
                            <xsl:message>
                                <xsl:value-of select="concat('WARNING: `keyfile` appoints the default',
                                                             ' path, however that granularity not',
                                                             ' supported, will enable encryption but',
                                                             ' overwrite that location at the nodes')"/>
                            </xsl:message>
                        </xsl:when>
                        <xsl:when test="@key">
                            <xsl:message>
                                <xsl:value-of select="concat('WARNING: `key` granularity not',
                                                             ' supported, will at least enable',
                                                             ' encryption as such')"/>
                            </xsl:message>
                        </xsl:when>
                    </xsl:choose>
                    <xsl:choose>
                        <xsl:when test="$pcscmd_extra_corosync_encryption_forced">
                            <xsl:message>
                                <xsl:value-of select="concat('NOTE: current pcs will enable encryption',
                                                             ' automatically, no intervention needed')"/>
                            </xsl:message>
                        </xsl:when>
                        <xsl:otherwise>
                            <xsl:value-of select="' --encryption 1'"/>
                        </xsl:otherwise>
                    </xsl:choose>
                </xsl:when>
                <xsl:otherwise>
                    <xsl:message>
                        <xsl:value-of select="concat('WARNING: encryption requested, but current pcs',
                                                     ' not capable to enforce it')"/>
                    </xsl:message>
                </xsl:otherwise>
            </xsl:choose>
        </xsl:when>
        <xsl:when test="$pcscmd_extra_corosync_encryption_forced">
            <xsl:message>
                <xsl:value-of select="concat('WARNING: encryption not requested, but current pcs',
                                             ' will enforce it regardless')"/>
            </xsl:message>
        </xsl:when>
    </xsl:choose>
'''



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

###

from ....filters._2pcscmd import verbose_ec_test, verbose_inform
from ....filters.cib2pcscmd import attrset_xsl
from ....utils_xslt import NL

cib2pcscmd = ('''\
    <!-- "pcs alert create" only supported with certain newer
         versions of pcs -->
    <xsl:variable name="AlertId" select="@id"/>
    <xsl:choose>
        <xsl:when test="$pcscmd_extra_alerts">
''' + (
            verbose_inform('"add alert listener: ", $AlertId')
) + '''
            <xsl:value-of select='concat($pcscmd_pcs, "alert create &apos;path=",
                                         @path, "&apos; id=", $AlertId)'/>
            <xsl:if test="@description">
                <xsl:value-of select='concat(" &apos;description=",
                                             @description, "&apos;")'/>
            </xsl:if>
''' + (
            attrset_xsl("instance_attributes", cmd='" options"')
            +
            attrset_xsl("meta_attributes", cmd='" meta"')
) + '''
            <xsl:value-of select="'%(NL)s'"/>
''' + (
            verbose_ec_test
) + '''
            <!-- select (CIB 2.10+)  -->
            <xsl:if test="select">
                <xsl:choose>
                    <xsl:when test="$pcscmd_extra_alerts_select">
                        <xsl:message>%(alerts_select_notimpl_msg)s</xsl:message>
                    </xsl:when>
                    <xsl:otherwise>
                        <xsl:message>%(alerts_select_msg)s</xsl:message>
                    </xsl:otherwise>
                </xsl:choose>
            </xsl:if>

            <!-- recipients -->
            <xsl:for-each select="recipient">
''' + (
                verbose_inform('"add recipient for alert listener: ",'
                               ' @id, " for ", $AlertId')
) + '''
                <xsl:value-of select='concat($pcscmd_pcs, "alert recipient add ",
                                             $AlertId,
                                             " &apos;value=", @value, "&apos;",
                                             " id=", @id)'/>
                <xsl:if test="@description">
                    <xsl:value-of select='concat(" &apos;description=",
                                                 @description, "&apos;")'/>
                </xsl:if>
''' + (
                attrset_xsl("instance_attributes", cmd='" options"')
                +
                attrset_xsl("meta_attributes", cmd='" meta"')
) + '''
                <xsl:value-of select="'%(NL)s'"/>
''' + (
                verbose_ec_test
) + '''
            </xsl:for-each>

        </xsl:when>
        <xsl:otherwise>
            <xsl:message>%(alerts_msg)s</xsl:message>
        </xsl:otherwise>
    </xsl:choose>
''') % dict(
    NL=NL,
    alerts_msg="WARNING: target pacemaker/pcs versions do not support"
               " alerts, hence omitted",
    alerts_select_msg="WARNING: target pacemaker/pcs versions do not support"
                      " 'select' construct for alerts, hence omitted",
    alerts_select_notimpl_msg="WARNING: waiting for pcs to honor 'select'"
                              " construct for alerts in its syntax"
                              " (https://bugzilla.redhat.com/1512707)",
)



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

###

from ....utils_2to3 import execfile
from ....utils_prog import dirname_x

from logging import getLogger
log = getLogger(__name__)

# XXX a bit dirty DRY approach
from os.path import join
use = join(dirname_x(__file__, 2), '_clone_master.py')
myglobals = dict(__package__=__package__, __name__=__name__)
try:
    execfile(use, myglobals)
except IOError:
    log.error("Unable to refer to `{0}' file".format(use))
else:
    cib2pcscmd = myglobals['cib2pcscmd']



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""cmd-annotate filter"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from .. import package_name, version
from ..filter import Filter
from ..utils import args2tuple
from ..utils_2to3 import bytes_enc

from datetime import datetime
from platform import python_implementation, python_version
from sys import argv


cmd_annotate_self_id = ' '.join((package_name(), version))


@Filter.deco('Nothing', 'string-iter')
def cmd_annotate(flt_ctxt, in_obj):
    """Emit a comment block with clufter version + command used + target info"""
    ret = (''.join(('#', l)) for l in (
        lambda s: () if not s else (s, ))(
            flt_ctxt.get('annotate_shell', '/bin/false').join('! ').rstrip(' !')
    ) + (
        " sequence generated on {0} with: {1}".format(
            datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            cmd_annotate_self_id,
        ),
        " invoked as: {0}".format(repr(argv)),
        " targeting system: {0}".format(
            repr(args2tuple(flt_ctxt['system'], *flt_ctxt['system_extra']))
        ),
        " using interpreter: {0}".format(' '.join((
            python_implementation(), python_version(),
        ))),
    ))
    return ('bytestringiter', (bytes_enc(l, 'utf-8') for l in ret))



# -*- coding: UTF-8 -*-
# Copyright 2015 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ....utils_cib import ResourceSpec, rg2hb_xsl

ccsflat2cibprelude = '''\
    <!--
        mysql ~ mysql
     -->
    <xsl:when test="name() = 'mysql'">
''' + (
        ResourceSpec('ocf:heartbeat:mysql').xsl_attrs
) + '''
        <!-- INSTANCE_ATTRIBUTES -->
        <instance_attributes id="{concat($Prefix, '-ATTRS')}">
''' + (
            rg2hb_xsl('config', 'config_file')
) + '''\
            <!-- additional_parameters ~ listen_address + mysqld_options -->
            <xsl:if test="@listen_address or @mysqld_options">
            <nvpair id="{concat($Prefix, '-ATTRS-additional_parameters')}"
                    name="additional_parameters">
                <xsl:attribute name="value">
                    <xsl:if test="@listen_address">
                        <xsl:value-of select="concat(
                            '--bind-address=',
                            @listen_address,
                            ' '
                        )"/>
                    </xsl:if>
                    <xsl:value-of select="@mysqld_options"/>
                </xsl:attribute>
            </nvpair>
            </xsl:if>
        </instance_attributes>

        <!-- OPERATIONS -->
        <operations>
''' + (
            rg2hb_xsl('start', 'startup_wait', op=True)
            +
            rg2hb_xsl('stop', 'shutdown_wait', op=True)
) + '''\
        </operations>
    </xsl:when>
'''

###

from ....filters.ccs_artefacts import artefact_cond_ra

ccs_artefacts = ''.join((
    artefact_cond_ra('@config_file',
                     kind='A', desc='configuration file'),
))



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""cib2pcscmd filter"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ..facts import infer
from ..filter import XMLFilter
from ..filters._2pcscmd import verbose_ec_test, verbose_inform
from ..utils_xslt import NL, xslt_params


def attrset_xsl(attrset, cmd=None, inform=None):
    return ('''\
    <xsl:if test="{attrset}/nvpair">
        <xsl:choose>
            <xsl:when test="{attrset}/rule">
                <xsl:message>
                    <!-- TODO:PCS -->
                    <xsl:value-of select="concat('WARNING: has to skip rule-based',
                                                ' {attrset} ', @id,
                                                ' (rhbz#1250744)')"/>
                </xsl:message>
            </xsl:when>
            <xsl:otherwise>
''' + (
                (verbose_inform(inform) + '\n' if inform else '')
                +
                ('''<xsl:value-of select='concat("", {cmd})'/>'''
                 if cmd else '')
) + '''
                <xsl:for-each select="{attrset}">
                    <xsl:for-each select="nvpair">
                        <xsl:value-of select='concat(" &apos;",
                                                    @name, "=", @value,
                                                    "&apos;")'/>
                    </xsl:for-each>
                </xsl:for-each>
''' + (
                ('''<xsl:value-of select="'{NL}'"/>''' + '\n'
                 + verbose_ec_test) if cmd and inform else ''
) + '''
            </xsl:otherwise>
        </xsl:choose>
    </xsl:if>
''').format(NL=NL, attrset=attrset, cmd=cmd)


@XMLFilter.deco('cib-2', 'string-list', defs=dict(
    pcscmd_force=False,
    pcscmd_verbose=True,
    pcscmd_tmpcib='tmp-cib.xml',
    pcscmd_dryrun=False,
))
def cib2pcscmd(flt_ctxt, in_obj):
    """Outputs set of pcs commands to reinstate the cluster per existing CIB"""
    self = flt_ctxt.ctxt_wrapped
    dry_run, tmp_cib = flt_ctxt['pcscmd_dryrun'], flt_ctxt['pcscmd_tmpcib']
    tmp_cib = (tmp_cib or self.defs['pcscmd_tmpcib']) if dry_run else tmp_cib
    return (
        'bytestring',
        flt_ctxt.ctxt_proceed_xslt(
            in_obj,
            textmode=True,
            def_first=xslt_params(
                pcscmd_force=flt_ctxt['pcscmd_force'],
                pcscmd_verbose=flt_ctxt['pcscmd_verbose'],
                pcscmd_tmpcib=flt_ctxt['pcscmd_tmpcib'],
                pcscmd_dryrun=dry_run,
                pcscmd_pcs="pcs -f {0} ".format(tmp_cib) if tmp_cib else "pcs ",

                pcscmd_extra_utilization = bool(infer(
                    'comp:pcs[utilization]',
                    flt_ctxt['system'],
                    flt_ctxt['system_extra'],
                )),
                pcscmd_extra_alerts = bool(infer(
                    'comp:pacemaker[alerts] + comp:pcs[alerts]',
                    flt_ctxt['system'],
                    flt_ctxt['system_extra'],
                )),
                pcscmd_extra_agents_via_pacemaker = bool(infer(
                    'comp:pcs[agents-via-pacemaker]',
                    flt_ctxt['system'],
                    flt_ctxt['system_extra'],
                )),
                pcscmd_extra_acls = bool(infer(
                    'comp:pacemaker[schema-2.0] + comp:pcs[acls]',
                    flt_ctxt['system'],
                    flt_ctxt['system_extra'],
                )),
                pcscmd_extra_push_diff  = bool(infer(
                    'comp:pcs[push-diff]',
                    flt_ctxt['system'],
                    flt_ctxt['system_extra'],
                )),
                pcscmd_extra_bundle = bool(infer(
                    'comp:pacemaker[bundle] + comp:resource-agents[docker] + comp:pcs[bundle]',
                    flt_ctxt['system'],
                    flt_ctxt['system_extra'],
                )),
                pcscmd_extra_bundle_meta = bool(infer(
                    'comp:pacemaker[bundle] + comp:pcs[bundle-meta]',
                    flt_ctxt['system'],
                    flt_ctxt['system_extra'],
                )),
                pcscmd_extra_alerts_select = bool(infer(
                    'comp:pacemaker[schema-2.10] + comp:pcs[alerts-select]',
                    flt_ctxt['system'],
                    flt_ctxt['system_extra'],
                )),
            ),
        ),
    )



# -*- coding: UTF-8 -*-
# Copyright 2014 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""ccs-subst-nodes filter"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ..filter import FilterError, XMLFilter
from ..utils_cman import get_nodes

from logging import getLogger

log = getLogger(__name__)

@XMLFilter.deco('ccs', 'ccs')
def ccs_subst_nodes(flt_ctxt, in_obj):
    # could be in-situ, but stick with the default assumptions;
    # we temporarily store also the old name as <OLD>@<NEW> as name
    from copy import deepcopy

    self = flt_ctxt.ctxt_wrapped

    ret = deepcopy(in_obj('etree'))
    declared, existing = flt_ctxt['nodes'].split(' '), get_nodes(ret)
    declared_no, existing_no = map(len, (declared, existing))
    if declared_no != existing_no:
        raise FilterError(self, "number of provided node names does not match"
                                " the real configuration: {0} vs {1}"
                                .format(declared_no, existing_no))
    declared.reverse()
    for clusternode in existing:
        old, new = clusternode.attrib['name'], declared.pop()
        clusternode.attrib['name'] = '@'.join((old, new))
        log.info("substituted nodename: {0} -> {1}".format(old, new))

    # pick <OLD>@<NEW> naming from previous phase and run some implied
    # translations, dropping "<OLD>@" part along
    return ('etree', flt_ctxt.ctxt_proceed_xslt(type(in_obj)('etree', ret)))



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

###

from ....filters._2pcscmd import verbose_ec_test, verbose_inform
from ....utils_xslt import NL, xslt_is_member

# http://clusterlabs.org/doc/en-US/Pacemaker/1.1-pcs/html-single/Pacemaker_Explained/#_cluster_options
cib2pcscmd_whitelist = (
    'no-quorum-policy',
    'batch-limit',
    'migration-limit',
    'symmetric-cluster',
    'stop-all-resources',
    'stop-orphan-resources',
    'stop-orphan-actions',
    'start-failure-is-fatal',
    'enable-startup-probes',
    'maintenance-mode',
    'stonith-enabled',
    'stonith-action',
    'stonith-timeout',
    'cluster-delay',
    'dc-deadtime',
    'cluster-recheck-interval',
    'pe-error-series-max',
    'pe-warn-series-max',
    'pe-input-series-max',
    'remove-after-stop',
    'startup-fencing',
    #'election-timeout',
    #'shutdown-escalation',
    #'crmd-integration-timeout',
    #'crmd-finalization-timeout',
    'crmd-transition-delay',
    #'default-resource-stickiness',
    #'is-managed-default',
    #'default-action-timeout',
)

# delay list a.k.a. what can play a role later on, but not now
cib2pcscmd_delaylist = (
    'enable-acl',
)

cib2pcscmd = ('''\
    <xsl:for-each select="cluster_property_set">
        <xsl:for-each select="nvpair">
            <!-- unfortunately pcs will throw away the rest of name-value
                 pairs when first unknown observed; alternatively a single
                 command with "force" -->
            <xsl:choose>
                <xsl:when test="
''' + (
                    xslt_is_member('@name', cib2pcscmd_whitelist)
) + '''">
''' + (
                    verbose_inform('"set singleton cluster property: ", @name')
) + '''
                    <xsl:value-of select='concat($pcscmd_pcs, "property set")'/>
                    <xsl:if test="$pcscmd_force">
                        <xsl:value-of select="' --force'"/>
                    </xsl:if>
                    <xsl:value-of select='concat(" &apos;", @name, "=", @value, "&apos;")'/>
                    <xsl:value-of select="'%(NL)s'"/>
''' + (
                    verbose_ec_test
) + '''
                </xsl:when>
                <xsl:when test="
''' + (
                    xslt_is_member('@name', cib2pcscmd_delaylist)
) + '''">
                </xsl:when>
                <xsl:otherwise>
                    <xsl:message>
                        <xsl:value-of select="concat(
                                                  'WARNING: dropping non-whitelisted cluster property: `',
                                                   @name,
                                                   '`'
                                              )"/>
                    </xsl:message>
                </xsl:otherwise>
            </xsl:choose>
        </xsl:for-each>
    </xsl:for-each>
''') % dict(
    NL=NL,
)

###

from ....utils_2to3 import execfile
from ....utils_prog import dirname_x
from ....utils_xslt import xslt_is_member

from logging import getLogger
log = getLogger(__name__)

# XXX a bit dirty DRY approach
from os.path import join
use = join(dirname_x(__file__, 2), '__init__.py')
myglobals = dict(__package__=__package__, __name__=__name__)
try:
    execfile(use, myglobals)
except IOError:
    log.error("Unable to refer to `{0}' file".format(use))
    cib_revitalize_deprecated_props_cluster = None  # make it fail later
else:
    cib_revitalize_deprecated_props_cluster = \
        myglobals['cib_revitalize_deprecated_props_cluster']

cib_revitalize = ('''\
    <xsl:copy>
        <xsl:for-each select="cluster_property_set">
            <xsl:if test="not(
                           count(nvpair[
''' + (
                                  xslt_is_member('@name',
                                                 cib_revitalize_deprecated_props_cluster)
) + '''
                              ]) = count(nvpair)
                          ) and not(@id-ref)">
                <xsl:copy>
                    <xsl:copy-of select="@*"/>
                    <xsl:apply-templates select="rule"/>
                    <xsl:apply-templates select="nvpair[not(
''' + (
                                                     xslt_is_member('@name',
                                                                    cib_revitalize_deprecated_props_cluster)
) + ''')]"/>
                    <xsl:apply-templates select="score"/>
                </xsl:copy>
            </xsl:if>
        </xsl:for-each>
    </xsl:copy>
''')



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""Easy (at least for usage) plugin framework"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from contextlib import contextmanager
from fnmatch import translate
from imp import PY_SOURCE, find_module, get_suffixes, load_module
from logging import getLogger
from os import pathsep, walk
from os.path import abspath, dirname, isabs, isdir, join, splitext
from re import compile as re_compile
from sys import modules

try:
    from .defaults import EXTPLUGINS_SHARED
except ImportError:
    EXTPLUGINS_SHARED = ''
from .utils import args2tuple, \
                   args2sgpl, \
                   classproperty, \
                   filterdict_keep, \
                   filterdict_pop, \
                   filterdict_remove, \
                   hybridproperty, \
                   tuplist
from .utils_2to3 import StandardError, filter_u, foreach_u  #, iter_values
from .utils_func import apply_intercalate
from .utils_prog import ProtectedDict, cli_decor, getenv_namespaced

log = getLogger(__name__)
module_ext = dict((t, s) for s, m, t in get_suffixes())[PY_SOURCE]

here = dirname(abspath(__file__))

EXTPLUGINS = tuple(e if isabs(e) else join(here, e) for e in
    map(str.strip, getenv_namespaced(
        'EXTPLUGINS', pathsep.join(('ext-plugins', EXTPLUGINS_SHARED))
    ).rstrip(pathsep).split(pathsep))
)


class MetaPlugin(object):
    """For use in the internal meta-hierarchy as "do not register me" mark"""


class PluginRegistry(type):
    """Core of simple plugin framework

    This ``superclass'' serves two roles:

    (1) metaclass for plugins (and its class hierarchy)
        - only a tiny wrapper on top of `type`

    (2) base class for particular registries

    Both roles are combined due to a very tight connection.

    Inspired by http://eli.thegreenplace.net/2012/08/07/ (Fundamental...).
    """
    _registries = set()  # dynamic tracking of specific registries

    #
    # these are relevant for use case (1)
    #

    # non-API

    def __new__(registry, name, bases, attrs):
        assert '_probes' not in attrs, "sabotage of the meta level detected"
        attrs['_probes'] = 0
        if '__metaclass__' not in attrs and MetaPlugin not in bases:
            # alleged end-use plugin
            ret = registry.probe(name, bases, attrs)
        else:
            if registry not in PluginRegistry._registries:
                log.debug("Registering `{0}' as registry"
                            .format(registry.registry))
                # specific registry not yet tracked
                # (e.g., specific plugin was imported natively)
                registry.setup()
                PluginRegistry._registries.add(registry)
                # rely on sys being always present (tautology) -> failing test
                if (registry.namespace or 'sys') not in modules:
                    # XXX hack to prevent RuntimeWarning due to missing parent
                    #     module (e.g., some tests in the test suite)
                    __import__(registry.namespace)

            ret = super(PluginRegistry, registry).__new__(registry, name,
                                                          bases, attrs)

        ret._probes += 1
        return ret

    #
    # these are relevant for both (1) + (2)
    #

    @classmethod
    def probe(registry, name, bases, attrs=None):
        """Meta-magic to register plugin"""
        assert '-' not in name, "name cannot contain a dash"
        dname = cli_decor(name)
        attrs = attrs or {}
        try:
            ret = registry._plugins[dname]
            log.info("Probe `{0}' plugin under `{1}' registry: already tracked"
                     .format(dname, registry.registry))
        except KeyError:
            log.debug("Probe `{0}' plugin under `{1}' registry: yet untracked"
                      .format(dname, registry.registry))
            namespaced = registry.namespaced(dname.split('-', 1)[0])
            if namespaced in modules and hasattr(modules[namespaced], name):
                # XXX can be used even for "preload"
                ret = getattr(modules[namespaced], name)
            else:  # bases arg of type has to be tuplist, final plugin otherwise
                ret = bases if not tuplist(bases) else \
                      super(PluginRegistry, registry).__new__(registry, dname,
                                                              bases, attrs)
            # XXX init plugin here?
            registry._plugins[dname] = ret

        if registry._path_context is not None:
            registry._path_mapping[registry._path_context].add(dname)
        if tuplist(bases) and ret.__metaclass__.__module__ == ret.__module__:
            ret.__metaclass__._native_plugins[cli_decor(ret.name)] = ret

        return ret

    @hybridproperty
    def name(this):
        """Nicer access to __name__"""
        return this.__name__

    @classproperty
    def registry(registry):
        """Registry identifier"""
        return registry.__name__

    @classproperty
    def namespace(registry):
        """Absolute namespace possessed by the particular plugin registry

        For a plugin, this denotes to which registry/namespace it belongs.
        """
        try:
            return registry._namespace
        except AttributeError:
            registry._namespace = '.'.join((__package__, registry.__name__))
            return registry._namespace

    @classmethod
    def namespaced(registry, symbol, *symbols):
        return '.'.join(args2tuple(registry.namespace, symbol, *symbols))

    @classproperty
    def plugins(registry):
        return registry._plugins_ro

    @classproperty
    def native_plugins(registry):
        return registry._native_plugins_ro

    #
    # these are relevant for use case (2)
    #

    # non-API

    @classmethod
    @contextmanager
    def _path(registry, path):
        """Temporary path context setup enabling safe sideways use"""
        assert registry._path_context is None
        registry._path_context = path
        yield path, registry._path_mapping.setdefault(path, set())
        assert registry._path_context is not None
        registry._path_context = None

    @classmethod
    def _context(registry, paths):
        """Iterate through the paths yielding context along

        Context is a pair `(path, list_of_per_path_tracked_plugins_so_far)`.
        """
        paths = args2sgpl(paths)
        if paths and paths[0] is None:  # injection of implicit path prevented
            paths = paths[1:]
        else:  # injection happens
            paths = args2tuple(dirname(abspath(__file__)), *paths)
        paths = (p for p in (join(p, registry.__name__) for p in paths)
                 if isdir(p))

        for path in paths:
            with registry._path(path) as context:
                yield context

    # API

    @classmethod
    def setup(registry, reset=False):
        """Implicit setup upon first registry involvement or external reset"""
        ps = {}
        attrs = (('_path_context', None), ('_path_mapping', {}),
                 ('_plugins', ps), ('_plugins_ro', ProtectedDict(ps)))
        if reset:
            foreach_u(lambda a, d: setattr(registry, a, d), attrs)
        else:
            foreach_u(lambda a, d: setattr(registry, a,
                                           getattr(registry, a, d)), attrs)
        registry._native_plugins = np = getattr(registry, '_native_plugins', {})
        registry._native_plugins_ro = ProtectedDict(np)
        if reset:
            PluginRegistry._registries.discard(registry)

    @classmethod
    def discover(registry, paths=(), fname_start='', from_scratch=False):
        """Find relevant plugins available at the specified path(s)

        Keyword arguments:
            fname_start     glob/fnmatch pattern (not RE) used to filter files,
                            can also be a tuplist of strings like this

        Returns `{plugin_name: plugin_cls}` mapping of plugins found.
        """
        ret = {}
        fname_start_use = apply_intercalate(args2sgpl(fname_start or '[!_.]'))
        fp = re_compile('|'.join(
            translate(fs + '*')
            for fs in (pfx.split('-', 1)[0] for pfx in fname_start_use)
        ))
        if from_scratch:
            registry.setup(True)
        for path, path_plugins in registry._context(paths):
            # skip if path already discovered (considered final)
            if not path_plugins:
                # visit *.py files within (and under) the path and probe them
                for root, dirs, files in walk(path):
                    if root != path:
                        break  # ATM we only support flat dir (nested ~ private)
                    #try:
                    #    dirs.remove('__pycache__')  # PY3 (if we ever nest)
                    #except ValueError:
                    #    pass
                    for name, ext in filter_u(
                        lambda name, ext:
                            fp.match(name) and
                            (ext == module_ext or isdir(join(root, name))),
                        (splitext(x) for x in files + dirs)
                    ):
                        try:
                            mfile, mpath, mdesc = find_module(name, [root])
                        except ImportError:
                            log.debug("Omitting `{0}' at `{1}'"
                                      .format(name, root))
                        else:
                            mname = registry.namespaced(name)
                            try:
                                load_module(mname, mfile, mpath, mdesc)
                            except StandardError as e:
                                # non-fatal, just log it and keep going
                                log.error("Module load error: {0}: {1}"
                                          .format(mfile or mpath, str(e)))
                            finally:
                                if mfile:
                                    mfile.close()
                path_plugins = registry._path_mapping[path]
                if fname_start:  # not picking everything -> restart next time
                    registry._path_mapping.pop(path)

            # filled as a side-effect of meta-magic, see `__new__`
            ret.update((n, registry._plugins[n]) for n in path_plugins)

        return ret


class PluginManager(object):
    """Common (abstract) base for *Manager objects"""

    _default_registry = PluginRegistry

    @classmethod
    def lookup(cls, plugins, registry=None, **kwargs):
        ret, to_discover = {}, set()
        registry = cls._default_registry if registry is None else registry
        for plugin in args2sgpl(plugins):
            # XXX we could introspect sys.modules here as well
            try:
                ret[plugin] = registry.plugins[plugin]
            except KeyError:
                to_discover.add(plugin)
        kwargs = filterdict_keep(kwargs, 'paths', 'from_scratch')
        ret.update(registry.discover(fname_start=tuple(to_discover), **kwargs))

        to_discover.difference_update(ret)
        native_plugins = registry.native_plugins
        ret.update(filterdict_remove(to_discover,
                                     _fn_=lambda x: native_plugins[x],
                                     *native_plugins.keys()))
        to_discover = apply_intercalate(tuple(to_discover))
        if to_discover:
            log.debug("Couldn't look up everything: {0}".format(', '.join(
                                                                to_discover)))
        return ret  # if tuplist(plugins) else ret[plugins]

    @classmethod
    def init_lookup(cls, plugin=(), *plugins, **kwargs):
        plugins = args2sgpl(plugin, *plugins)
        ext_plugins = []
        if kwargs.get('ext_plugins', True):
            for e in EXTPLUGINS:
                for root, dirs, _ in walk(e, followlinks=True):
                    ext_plugins.extend(join(root, d) for d in dirs)
                    break
        ext_plugins.extend(
            abspath(d) for d in
            apply_intercalate([d.split(':') for d in
                kwargs.get('ext_plugins_user') or ()
            ])
        )
        if ext_plugins:
            if not isinstance(kwargs.setdefault('paths', []), list):
                kwargs['paths'] = paths = list(kwargs['paths'])
            else:
                paths = kwargs['paths']
            paths.extend(ext_plugins)
        kws_lu = filterdict_pop(kwargs, 'paths')
        return cls(plugins=cls.lookup(plugins, **kws_lu), paths=None, **kwargs)

    def __init__(self, *args, **kwargs):
        registry = kwargs.pop('registry', None) or self._default_registry
        assert registry is not PluginRegistry, \
               "PluginManager subclass should refer to its custom registry"
        self._registry = registry
        plugins = registry.discover(**filterdict_pop(kwargs, 'paths'))
        plugins.update(kwargs.pop('plugins', None)
                       or dict(registry.native_plugins, **registry.plugins))
        self._plugins = ProtectedDict(
            self._init_plugins(plugins, *args, **kwargs),
        )

    @classmethod
    def _init_plugins(cls, plugins, *args, **kwargs):
        log.info("Plugins under `{0}' left intact: {1}".format(cls.__name__,
                                                               plugins))
        return plugins

    @property
    def registry(self):
        return self._registry

    @property
    def plugins(self):
        return self._plugins

    #def __iter__(self):
    #    return iter_values(self._plugins)



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)

from __future__ import print_function

"""xml2simpleconfig filter"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ..filter import XMLFilter
from ..utils_2to3 import iter_items
from lxml import etree


@XMLFilter.deco('XML', 'simpleconfig')
def xml2simpleconfig(flt_ctxt, in_obj):
    """Mapping (almost bijective) XML -> simpleconfig

    Inverse mapping cannot be generally loseless, as XML cannot contain
    repeated attributes, which seems/is valid with simpleconfig.
    See `simpleconfig` docstring for details about the target representation.
    """
    # using similar trick of stack emulation in-place as command.analyse_chain.
    # but this is iterative rather than recursive :)
    root = []
    for action, e in etree.iterwalk(in_obj('etree'), events=('start', 'end')):
        if action == 'start':
            #print(">>> start", e.tag, root)
            current = [e.tag, None, None]
            root.append(current)
            current[1] = tuple(sorted(iter_items(e.attrib)))
            #print("<<< start", e.tag, root)
        elif action == 'end':
            #print(">>> end", e.tag, root)
            if len(root) == 1:
                #assert id(e) == id(root)
                break
            current = root.pop()
            if root[-1][2] is None:
                root[-1][2] = []
            root[-1][2].append(current)
            #print("<<< end", e.tag, root)
    return ('struct', root[-1])



# -*- coding: UTF-8 -*-
# Copyright 2016 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""ccspcmk2pcscmd filter"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ..facts import infer
from ..filter import XMLFilter
from ..utils_xslt import xslt_params


@XMLFilter.deco('ccs', 'string-list', defs=dict(
    pcscmd_force=False,
    pcscmd_noauth=False,
    pcscmd_verbose=True,
    pcscmd_dryrun=False,
    pcscmd_enable=False,
    pcscmd_start_wait=60,
    pcscmd_noguidance=0,
))
def ccspcmk2pcscmd(flt_ctxt, in_obj):
    """Outputs set of pcs commands to reinstate the cluster per cluster.conf"""
    return (
        'bytestring',
        flt_ctxt.ctxt_proceed_xslt(
            in_obj,
            textmode=True,
            def_first=xslt_params(
                pcscmd_force=flt_ctxt['pcscmd_force'],
                pcscmd_noauth=flt_ctxt['pcscmd_noauth'],
                pcscmd_verbose=flt_ctxt['pcscmd_verbose'],
                pcscmd_dryrun=flt_ctxt['pcscmd_dryrun'],
                pcscmd_enable=flt_ctxt['pcscmd_enable'],
                pcscmd_start_wait=flt_ctxt['pcscmd_start_wait'],
                pcscmd_noguidance=flt_ctxt['pcscmd_noguidance'],

                pcscmd_extra_wait_cluster_start = bool(infer(
                    'comp:pcs[wait-cluster-start]',
                    flt_ctxt['system'],
                    flt_ctxt['system_extra'],
                )),
            ),
        ),
    )



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

cib2pcscmd = '''\
    <clufter:descent-mix at="acl_role"/>
    <clufter:descent-mix at="acl_target"/>
    <clufter:descent-mix at="acl_group"/>
'''



# -*- coding: UTF-8 -*-
# Copyright 2015 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)

###

needlexml2pcscmd = '''\
    <xsl:variable name="RingNumber">
        <xsl:choose>
            <xsl:when test="@ringnumber">
                <xsl:value-of select="@ringnumber"/>
            </xsl:when>
            <xsl:otherwise>
                <xsl:value-of select="count(
                    preceding-sibling::interface[not(@ringnumber)]
                )"/>
            </xsl:otherwise>
        </xsl:choose>
    </xsl:variable>
    <xsl:if test="$RingNumber &gt;= 0 or $RingNumber &lt;= 1">
        <xsl:if test="@bindnetaddr">
            <xsl:value-of select="concat(' --addr', $RingNumber, ' ', @bindnetaddr)"/>
        </xsl:if>
        <xsl:choose>
            <xsl:when test="@broadcast = 'yes'">
                <xsl:value-of select="concat(' --broadcast', $RingNumber)"/>
            </xsl:when>
            <xsl:otherwise>
                <xsl:if test="@mcastaddr">
                    <xsl:value-of select="concat(' --mcast', $RingNumber, ' ', @mcastaddr)"/>
                </xsl:if>
                <xsl:if test="@mcastport">
                    <xsl:value-of select="concat(' --mcastport', $RingNumber, ' ', @mcastport)"/>
                </xsl:if>
                <xsl:if test="@ttl">
                    <xsl:value-of select="concat(' --ttl', $RingNumber, ' ', @ttl)"/>
                </xsl:if>
            </xsl:otherwise>
        </xsl:choose>
    </xsl:if>
'''



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""needleqdevicexml2pcscmd filter"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ..facts import infer
from ..filter import XMLFilter
from ..utils_xslt import xslt_params


@XMLFilter.deco('coroxml-needle', 'string-list', defs=dict(
    pcscmd_force=False,
    pcscmd_verbose=True,
    pcscmd_noguidance=0,
))
def needleqdevicexml2pcscmd(flt_ctxt, in_obj):
    """Outputs set of pcs commands to reinstate the qdevice per corosync.conf"""
    return (
        'bytestring',
        flt_ctxt.ctxt_proceed_xslt(
            in_obj,
            textmode=True,
            def_first=xslt_params(
                pcscmd_force=flt_ctxt['pcscmd_force'],
                pcscmd_verbose=flt_ctxt['pcscmd_verbose'],
                pcscmd_noguidance=flt_ctxt['pcscmd_noguidance'],

                pcscmd_extra_qdevice = bool(infer(
                    'comp:corosync[qdevice] + comp:pcs[qdevice]',
                    flt_ctxt['system'],
                    flt_ctxt['system_extra'],
                )),
                pcscmd_extra_qdevice_heuristics = bool(infer(
                    'comp:corosync[qdevice-heuristics] + comp:pcs[qdevice-heuristics]',
                    flt_ctxt['system'],
                    flt_ctxt['system_extra'],
                )),
                pcscmd_extra_qnet = bool(infer(
                    'comp:corosync[qnet] + comp:pcs[qnet]',
                    flt_ctxt['system'],
                    flt_ctxt['system_extra'],
                )),
            ),
        ),
    )



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ....filters._2pcscmd import verbose_ec_test, verbose_inform
from ....utils_xslt import NL, xslt_is_member

# https://bugzilla.redhat.com/show_bug.cgi?id=1441332#c5
cib2pcscmd_id_whitelist = (
    'acl_role',
    'rsc_colocation',
    'rsc_location',
    'rsc_order',
    'rsc_ticket',
)

cib2pcscmd = ('''\
    <!-- "pcs acl" only supported with certain newer versions of
         pacemaker/pcs (https://bugzilla.redhat.com/1111369) -->
    <xsl:choose>
        <xsl:when test="$pcscmd_extra_acls">
''' + (
            verbose_inform('"new ACL role: ", @id')
) + '''
            <xsl:value-of select="concat($pcscmd_pcs, 'acl role create', ' ', @id)"/>
            <xsl:if test="@description != ''">
                <xsl:value-of select='concat(" &apos;description=", @description,
                                             "&apos;")'/>
            </xsl:if>
            <xsl:for-each select="acl_permission">
                <!-- XXX only allow a small amount of permissions to be in-line -->
                <xsl:choose>
                    <xsl:when test="@xpath">
                        <xsl:value-of select="concat(' ', @kind)"/>
                        <xsl:value-of select='concat(" xpath &apos;", @xpath, "&apos;")'/>
                    </xsl:when>
                    <xsl:when test="@reference">
                        <xsl:variable name="CibAclReferenced" select="//*[@id = current()/@reference]"/>
                        <xsl:choose>
                            <xsl:when test="
''' + (
                                xslt_is_member('name($CibAclReferenced)', cib2pcscmd_id_whitelist)
) + '''">
                            <xsl:value-of select="concat(' ', @kind)"/>
                                <xsl:value-of select="concat(' id ', @reference)"/>
                            </xsl:when>
                            <xsl:otherwise>
                                <xsl:message>
                                    <xsl:value-of select="concat('WARNING: ACL: current pcs',
                                                                 ' does not allow one to',
                                                                 ' specify identifier with `',
                                                                 name($CibAclReferenced),
                                                                 '` element')"/>
                                </xsl:message>
                                <xsl:value-of select="concat(' ', @kind)"/>
                                <xsl:value-of select="concat(' id ', @reference)"/>
                            </xsl:otherwise>
                        </xsl:choose>
                    </xsl:when>
                    <xsl:when test="@object-type">
                        <xsl:message>
                            <xsl:value-of select="concat('NOTE: ACL: current pcs can only',
                                                         ' express particular tag (optionally',
                                                         ' qualified with an attribute',
                                                         ' presence) as a subject of ACL',
                                                         ' within CIB indirectly with',
                                                         ' more complex XPath expression',
                                                         ' [https://bugzilla.redhat.com/1440890]')"/>
                        </xsl:message>
                        <xsl:value-of select="concat(' ', @kind)"/>
                        <xsl:value-of select='concat(" xpath &apos;//", @object-type)'/>
                        <xsl:if test="@attribute">
                            <xsl:value-of select="concat('[', @attribute, ']')"/>
                        </xsl:if>
                        <xsl:value-of select='"&apos;"'/>
                    </xsl:when>
                </xsl:choose>
            </xsl:for-each>
            <xsl:value-of select="'%(NL)s'"/>
''' + (
            verbose_ec_test
) + '''
        </xsl:when>
        <xsl:when test="not(preceding-sibling::acl_role)">
            <xsl:message>%(acls_msg)s</xsl:message>
        </xsl:when>
    </xsl:choose>
''') % dict(
    NL=NL,
    acls_msg="WARNING: target pacemaker/pcs versions do not support"
             " (new) ACLs, hence &apos;acl_role&apos; element(s) omitted",
)



# -*- coding: UTF-8 -*-
# Copyright 2015 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""ccs-artefacts command"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ..command import Command
from ..protocol import protocols
from ..utils_cman import PATH_CLUSTERCONF


@Command.deco('ccs-artefacts')
def ccs_artefacts(cmd_ctxt,
                  input=PATH_CLUSTERCONF,
                  output="cman-artefacts-{ccs-artefacts.in.hash}.conf"):
    """Output artefacts referenced in the config. in CVS format

    Options:
        input   input CMAN-based cluster configuration file
        output  output file with collected artefacts (files, etc.)
    """

    file_proto = protocols.plugins['file'].ensure_proto
    return (
        file_proto(input),
        file_proto(output),
    )
